#!/usr/bin/env python
'''
Created by Seria at 02/11/2018 3:38 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
import numpy as np
import random as rand
from collections import abc
from torchvision.transforms import *
F = functional
from PIL import Image

from ..toolkit import byte2arr


__all__ = ('Comburant', 'HWC2CHW', 'Random',
           'NEAREST', 'LINEAR', 'CUBIC', 'HORIZONTAL', 'VERTICAL',
           'Resize', 'Crop', 'Flip', 'Rotate',
           'Brighten', 'Contrast', 'Saturate', 'Hue')


NEAREST = 0
LINEAR = 1
CUBIC = 2

PIL_INTERP = {NEAREST: Image.NEAREST, LINEAR: Image.BILINEAR, CUBIC: Image.BICUBIC}

HORIZONTAL = 10
VERTICAL = 11




class Comburant(object):
    def __init__(self, *args, is_encoded=False):
        if isinstance(args[-1], HWC2CHW):
            ls_args = list(args[:-1])
            self.cvt_form = args[-1]
        else:
            ls_args = list(args)
        self.comburant = Compose(ls_args)
        self.is_encoded = is_encoded

    def __call__(self, imgs):
        if self.is_encoded:
            if isinstance(imgs, abc.Sequence):
                img = []
                for i in imgs:
                    img.append(byte2arr(i, as_np=False))
            else:
                img = byte2arr(imgs, as_np=False)
            imgs = img
        imgs = self.comburant(imgs)
        if isinstance(imgs, abc.Sequence):
            img = []
            for i in imgs:
                i = np.array(i)
                i = i.astype(np.float32) / 255
                if hasattr(self, 'cvt_form'):
                    i = self.cvt_form(i)
                img.append(i)
        else:
            img = np.array(imgs)
            img = img.astype(np.float32) / 255
            if hasattr(self, 'cvt_form'):
                img= self.cvt_form(img)
        return img



class ABC(object):
    def __init__(self):
        pass

    def call(self, *args, **kwargs):
        raise NotImplementedError

    def __call__(self, imgs):
        if isinstance(imgs, abc.Sequence):
            ret = []
            for i in imgs:
                ret.append(self.call(i))
        else:
            ret = self.call(imgs)

        return ret



class HWC2CHW(ABC):
    def __init__(self):
        super(HWC2CHW, self).__init__()

    def __call__(self, img):
        return np.transpose(img, (2, 0, 1))



class Random(ABC):
    def __init__(self, p, comburant):
        super(Random, self).__init__()
        self.p = p
        self.cbr = comburant

    def call(self, img, conduct):
        if conduct:
            return self.cbr(img)
        else:
            return img

    def __call__(self, imgs):
        if rand.random() < self.p:
            conduct = True
        else:
            conduct = False
        if isinstance(imgs, abc.Sequence):
            ret = []
            for i in imgs:
                ret.append(self.call(i, conduct))
        else:
            ret = self.call(imgs, conduct)

        return ret



class Resize(ABC):
    def __init__(self, size, interp=LINEAR):
        # size: (height, width)
        super(Resize, self).__init__()
        self.size = size
        self.interp = interp

    def call(self, img):
        return img.resize((self.size[1], self.size[0]), PIL_INTERP[self.interp])



class Crop(ABC):
    def __init__(self, size, padding=(0, 0, 0, 0), area_ratio=(1, 1), aspect_ratio=(1, 1), interp=LINEAR, scale=()):
        # size: (height, width)
        # padding: (left, top, right, bottom)
        super(Crop, self).__init__()
        self.size = size
        self.padding = padding
        self.area_ratio = area_ratio
        self.aspect_ratio = aspect_ratio
        self.scale = scale
        if area_ratio == aspect_ratio == (1,1):
            self.reshape = False
            self.comburant = RandomCrop(size)
        else:
            self.reshape = True
            self.comburant = RandomResizedCrop(size, area_ratio, aspect_ratio, PIL_INTERP[interp])

    def call(self, img, param, t=1):
        param = [p * t for p in param]
        y, x, h, w = param
        padding = tuple([p * t for p in self.padding])
        size = tuple([s * t for s in self.size])
        img = F.pad(img, padding, 0, 'constant')
        # pad the width if needed
        if img.size[0] < size[1]:
            img = F.pad(img, (size[1] - img.size[0], 0), 0, 'constant')
        # pad the height if needed
        if img.size[1] < size[0]:
            img = F.pad(img, (0, size[0] - img.size[1]), 0, 'constant')

        if self.reshape:
            return F.resized_crop(img, y, x, h, w, size, self.comburant.interpolation)
        else:
            return F.crop(img, y, x, h, w)


    def __call__(self, imgs):
        if len(self.scale) == 0:
            self.scale = len(imgs) * [1]
        img = F.pad(imgs[0], self.padding, 0, 'constant')
        # pad the width if needed
        if img.size[0] < self.size[1]:
            img = F.pad(img, (self.size[1] - img.size[0], 0), 0, 'constant')
        # pad the height if needed
        if img.size[1] < self.size[0]:
            img = F.pad(img, (0, self.size[0] - img.size[1]), 0, 'constant')
        if self.reshape:
            param = self.comburant.get_params(img, self.comburant.scale, self.comburant.ratio)
        else:
            param = self.comburant.get_params(img, self.comburant.size)

        if isinstance(imgs, abc.Sequence):
            ret = []
            for i, v in enumerate(imgs):
                ret.append(self.call(v, param, self.scale[i]))
        else:
            ret = self.call(imgs, param)

        return ret




class Flip(ABC):
    def __init__(self, axial):
        super(Flip, self).__init__()
        if axial == HORIZONTAL:
            self.comburant = RandomVerticalFlip(1)
        elif axial == VERTICAL:
            self.comburant = RandomHorizontalFlip(1)
        else:
            raise Exception('NEBULAE ERROR ⨷ the invoked flip type is not defined or supported.')

    def call(self, img):
        return self.comburant(img)



class Rotate(ABC):
    def __init__(self, degree, intact=False, interp=NEAREST):
        '''
        Args
        intact: whether to keep image intact which might enlarge the output size
        '''
        super(Rotate, self).__init__()
        self.comburant = RandomRotation(degree, PIL_INTERP[interp], intact)

    def call(self, img, angle):
        return F.rotate(img, angle, self.comburant.resample, self.comburant.expand,
                        self.comburant.center, self.comburant.fill)

    def __call__(self, imgs):
        angle = self.comburant.get_params(self.comburant.degrees)
        if isinstance(imgs, abc.Sequence):
            ret = []
            for i in imgs:
                ret.append(self.call(i, angle))
        else:
            ret = self.call(imgs, angle)

        return ret



class Brighten(ABC):
    def __init__(self, range):
        super(Brighten, self).__init__()
        self.comburant = ColorJitter(brightness=range)

    def call(self, img, factor):
        return F.adjust_brightness(img, factor)

    def __call__(self, imgs):
        factor = rand.uniform(self.comburant.brightness[0], self.comburant.brightness[1])
        if isinstance(imgs, abc.Sequence):
            ret = []
            for i in imgs:
                ret.append(self.call(i, factor))
        else:
            ret = self.call(imgs, factor)

        return ret



class Contrast(ABC):
    def __init__(self, range):
        super(Contrast, self).__init__()
        self.comburant = ColorJitter(contrast=range)

    def call(self, img, factor):
        return F.adjust_contrast(img, factor)

    def __call__(self, imgs):
        factor = rand.uniform(self.comburant.contrast[0], self.comburant.contrast[1])
        if isinstance(imgs, abc.Sequence):
            ret = []
            for i in imgs:
                ret.append(self.call(i, factor))
        else:
            ret = self.call(imgs, factor)

        return ret



class Saturate(ABC):
    def __init__(self, range):
        super(Saturate, self).__init__()
        self.comburant = ColorJitter(saturation=range)

    def call(self, img, factor):
        return F.adjust_saturation(img, factor)

    def __call__(self, imgs):
        factor = rand.uniform(self.comburant.saturation[0], self.comburant.saturation[1])
        if isinstance(imgs, abc.Sequence):
            ret = []
            for i in imgs:
                ret.append(self.call(i, factor))
        else:
            ret = self.call(imgs, factor)

        return ret



class Hue(ABC):
    def __init__(self, range):
        super(Hue, self).__init__()
        self.comburant = ColorJitter(hue=range)

    def call(self, img, factor):
        return F.adjust_hue(img, factor)

    def __call__(self, imgs):
        factor = rand.uniform(self.comburant.hue[0], self.comburant.hue[1])
        if isinstance(imgs, abc.Sequence):
            ret = []
            for i in imgs:
                ret.append(self.call(i, factor))
        else:
            ret = self.call(imgs, factor)

        return ret


#!/usr/bin/env python
'''
__init__
Created by Seria at 20/12/2018 4:54 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
from .engine import Engine
from .time_machine import TimeMachine
import os

core = os.environ.get('NEB_CORE', 'PYTORCH')
if core.upper() == 'TENSORFLOW':
    from .multiverse_tf import Multiverse
elif core.upper() == 'PYTORCH':
    from .multiverse_pt import Multiverse
else:
    raise ValueError('NEBULAE ERROR ⨷ %s is an unsupported core.' % core)


#!/usr/bin/env python
'''
engine_mx
Created by Seria at 12/02/2019 3:45 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========

'''
# -*- coding:utf-8 -*-
import tensorflow as tf
import os

class Multiverse(object):
    '''
    Args:
    nworld: world size
    '''

    def __init__(self, universe, nworld=1):
        self.universe = universe
        self.nworld = nworld
        self.rank = -1
        os.environ["WORLD_SIZE"] = str(nworld)

    def __call__(self):
        self.universe(self)

    def init(self):
        pass

    def scope(self):
        self.stg = tf.distribute.MirroredStrategy()
        return self.stg.scope()

    def Executor(self, func):
        @tf.function
        def _execWrapper(*args, **kwargs):
            return self.stg.run(func, args, kwargs)

        return _execWrapper

    def _sync(self, datum):
        datum.tdata = self.stg.experimental_distribute_dataset(datum.tdata)
        return datum

    def sync(self, models, data):
        if not isinstance(models, (list, tuple)):
            models = (models,)
        if not isinstance(data, (list, tuple)):
            data = (data,)

        synced_data = []
        for d in data:
            synced_data.append(self._sync(d))

        return tuple(models) + tuple(synced_data)

    def reduce(self, tensor, aggregate=False):
        if aggregate:
            reduce_op = tf.distribute.ReduceOp.SUM
        else:
            reduce_op = tf.distribute.ReduceOp.MEAN
        return self.stg.reduce(reduce_op, tensor, None)


#!/usr/bin/env python
'''
control_panel
Created by Seria at 2020/10/22 10:48 AM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
import sys

try:
    import termios
    from pynput import keyboard

    class HotKey(keyboard.HotKey):
        def __init__(self, keys, on_activate):
            super(HotKey, self).__init__(keys, on_activate)

        def press(self, key):
            """Updates the hotkey state for a pressed key.

            If the key is not currently pressed, but is the last key for the full
            combination, the activation callback will be invoked.

            Please note that the callback will only be invoked once.

            :param key: The key being pressed.
            :type key: Key or KeyCode
            """
            if key in self._keys and key not in self._state:
                self._state.add(key)
                if self._state == self._keys:
                    ret = self._on_activate()
                    return ret



    def switch_mode():
        return 0
    def show_curve_1():
        return 1
    def show_curve_2():
        return 2
    def show_curve_3():
        return 3
    def show_curve_4():
        return 4
    def show_curve_5():
        return 5
    def show_curve_6():
        return 6
    def show_curve_7():
        return 7

    def on_quit():
        termios.tcflush(sys.stdin, termios.TCIFLUSH)
        return False

    class CtrlPanel(keyboard.Listener):
        """A keyboard listener supporting a number of global hotkeys.

        This is a convenience wrapper to simplify registering a number of global
        hotkeys.

        :param dict hotkeys: A mapping from hotkey description to hotkey action.
            Keys are strings passed to :meth:`HotKey.parse`.

        :raises ValueError: if any hotkey description is invalid
        """
        def __init__(self, dashboard):
            self.rank = dashboard.param['rank']
            self.db = dashboard
            self.kc = keyboard.Controller()
            if self.rank>0:
                hotkeys = {}
            else:
                hotkeys = {'<ctrl>+0': switch_mode,
                           '<ctrl>+1': show_curve_1,
                           '<ctrl>+2': show_curve_2,
                           '<ctrl>+3': show_curve_3,
                           '<ctrl>+4': show_curve_4,
                           '<ctrl>+5': show_curve_5,
                           '<ctrl>+6': show_curve_6,
                           '<ctrl>+7': show_curve_7,
                           '<backspace>': on_quit}
            self._hotkeys = [
                HotKey(HotKey.parse(key), value)
                for key, value in hotkeys.items()]
            super(CtrlPanel, self).__init__(
                on_press=self._on_press,
                on_release=self._on_release,)

        def actuate(self):
            if self.rank > 0:
                return
            fd = sys.stdin
            old = termios.tcgetattr(fd)
            new = termios.tcgetattr(fd)
            new[3] = new[3] & ~termios.ECHO
            try:
                termios.tcsetattr(fd, termios.TCSADRAIN, new)
                self.kc.press(keyboard.Key.backspace)
                self.kc.release(keyboard.Key.backspace)
                self.join()
            finally:
                termios.tcsetattr(fd, termios.TCSADRAIN, old)

        def refresh(self):
            self.kc.press(keyboard.Key.ctrl)
            self.kc.press('0')
            self.kc.release(keyboard.Key.ctrl)
            self.kc.release('0')

            self.kc.press(keyboard.Key.ctrl)
            self.kc.press('0')
            self.kc.release(keyboard.Key.ctrl)
            self.kc.release('0')

        def _on_press(self, key):
            """The press callback.

            This is automatically registered upon creation.

            :param key: The key provided by the base class.
            """
            for hotkey in self._hotkeys:
                ret = hotkey.press(self.canonical(key))
                if ret is False:
                    return ret
                elif isinstance(ret, int):
                    if ret>0:
                        self.db.panel = ret-1
                    else:
                        self.db.is_global = not self.db.is_global

        def _on_release(self, key):
            """The release callback.

            This is automatically registered upon creation.

            :param key: The key provided by the base class.
            """
            for hotkey in self._hotkeys:
                _ = hotkey.release(self.canonical(key))

except:
    CtrlPanel = None


#!/usr/bin/env python
'''
component_tf
Created by Seria at 05/02/2019 1:41 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
from types import MethodType
import tensorflow as tf
import tensorflow_addons as tfa
import tensorflow_probability as tfp
from tensorflow.keras import Model, layers, losses, regularizers, optimizers, backend
from ..cockpit.engine_tf import EngineTF

# 'GraphConvEig', 'GraphConvLap'
# 'CBN', 'IN', 'CIN', 'LN', 'SN',
# 'AdaBelief'
__all__ = ('Craft', 'Rudder', 'Prober', 'Nozzle',
           'NEAREST', 'LINEAR', 'CUBIC',
           'Void', 'XavierNorm', 'XavierUnif', 'Normal', 'Uniform', 'Orthog', 'Zeros', 'Ones',
           'Conv', 'TransConv', 'GraphConvAdj', 'Dense', 'Embed', 'Identity',
           'Mean', 'Max', 'Min', 'Sum', 'Dot',
           'QR', 'Eigen', 'MaxEigen',
           'Reshape', 'Permute', 'Upscale', 'MaxPool', 'AvgPool',
           'EMA', 'Retroact',
           'Sqrt',
           'Concat',
           'D2S', 'S2D', 'SpDot',
           'Clip', 'Dropout', 'BN',
           'Relu', 'LRelu', 'Tanh', 'Sigm', 'Sftm', 'Sftp',
           'MAE', 'MSE', 'SigmXE', 'SftmXE',
           'AccCls',
           'StepLR', 'PolyLR', 'CosLR', 'ExpLR', 'WavyLR',
           'Momentum', 'Nesterov', 'RMSProp', 'Adam',)



NEAREST = 0
LINEAR = 1
CUBIC = 2
TF_INTERP = {NEAREST: 'nearest', LINEAR: 'bilinear', CUBIC: 'bicubic'}




class Craft(Model):
    def __init__(self, scope):
        super(Craft, self).__init__()
        self.scope = scope
        self.__pods = []
        self.__dict = {}
        self.__formulated = False

    def run(self, *args, **kwargs):
        raise NotImplementedError

    # @tf.function
    def _run(self, *args, **kwargs):
        return self.run(*args, **kwargs)

    def call(self, *args, **kwargs):
        # assert all([isinstance(a, Tensor) for a in args])
        kwargs.pop('training')
        # if hasattr(self, 'exec'):
        #     return self.exec(*args, **kwargs)
        return self._run(*args, **kwargs)

    def gear(self, gr):
        if isinstance(gr, bool):
            self.trainable = gr
        elif isinstance(gr, tf.python.eager.backprop.GradientTape):
            self._rudder = gr
            self.trainable = True
        elif isinstance(gr, EngineTF):
            pass
        else:
            raise Exception('NEBULAE ERROR ⨷ %s is not a valid type of gear.' % type(gr))
        return self

    def vars(self):
        return self.trainable_weights

    def dissect(self, *dummy_args, **dummy_kwargs):
        if 'export_path' in dummy_kwargs.keys():
            export_path = dummy_kwargs['export_path']
            dummy_kwargs.pop('export_path')
        else:
            export_path = ''
        # feed forward once
        _ = self(*dummy_args, **dummy_kwargs)

        nbytes = {tf.int8: 1, tf.int64: 8, tf.float16: 2, tf.float32: 4, tf.float64: 8}
        parambytes = sum([p.shape.num_elements() * nbytes[p.dtype] for p in self.vars()])
        if parambytes<1024:
            parambytes = '%6d B  ' % parambytes
        elif parambytes<1048576:
            parambytes = '%6.2f KB ' % (parambytes / 1024)
        elif parambytes<1073741824:
            parambytes = '%6.2f MiB' % (parambytes / 1048576)
        else:
            parambytes = '%6.2f GB ' % (parambytes / 1073741824)
        print('+' + (len(self.scope) + 26) * '-' + '+')
        print('| Craft-%s weighs %s |' % (self.scope, parambytes))
        print('+' + (len(self.scope) + 26) * '-' + '+')

        # convert to onnx
        # TODO: draw architectures
        if export_path:
            import tf2onnx
            import logging
            logging.basicConfig(level=logging.ERROR)

            spec = []
            cfnames = []
            for d in dummy_args + tuple(dummy_kwargs.values()):
                cfnames.append('inputs_%d' % len(spec))
                spec.append(tf.TensorSpec.from_tensor(d, name=cfnames[-1]))
            dform = tf.keras.backend.image_data_format()
            if dform=='channels_first':
                _ = tf2onnx.convert.from_keras(self, spec, opset=9, output_path=export_path, inputs_as_nchw=cfnames)
            else:
                _ = tf2onnx.convert.from_keras(self, spec, opset=9, output_path=export_path)

            logging.basicConfig(level=logging.WARNING)

    @property
    def pods(self):
        return self.__pods

    @pods.setter
    def pods(self, pods):
        if not self.__formulated:
            self.__pods = pods
            self.__formulated = True

    def __getitem__(self, key):
        paths = key.split('/')
        craft = self
        for p in paths[:-1]:
            craft = getattr(craft, p)
        if paths[-1] == '':
            return craft
        else:
            return craft.__dict[paths[-1]]

    def __setitem__(self, key, value):
        self.__dict[key] = value



class Rudder(tf.GradientTape):
    def __init__(self):
        super(Rudder, self).__init__()

class Prober(tf.GradientTape):
    def __init__(self):
        super(Prober, self).__init__(persistent=True)

class Nozzle(object):
    def __init__(self):
        pass

    def __enter__(self):
        return False

    def __exit__(self, *args):
        pass



# -------------------------------------- Layer --------------------------------------- #

class Void(object):
    def __init__(self):
        self.iniz = 'ConstantV2'

class XavierNorm(object):
    def __init__(self):
        self.iniz = 'GlorotNormalV2'

class XavierUnif(object):
    def __init__(self):
        self.iniz = 'GlorotUniformV2'

class Normal(object):
    def __init__(self):
        self.iniz = 'RandomNormalV2'

class Uniform(object):
    def __init__(self):
        self.iniz = 'RandomUniformV2'

class Orthog(object):
    def __init__(self):
        self.iniz = 'OrthogonalV2'

class Ones(object):
    def __init__(self):
        self.iniz = 'OnesV2'

class Zeros(object):
    def __init__(self):
        self.iniz = 'ZerosV2'



class Conv(Craft):
    def __init__(self, in_chs, out_chs, kernel: tuple, stride=1, padding=0, dilation=1, group=1,
                 w_init=XavierNorm(), b_init=Zeros(), scope='CONV'):
        '''
        Args:
        - in_chs: input channel
        - out_chs: output channel
        - kernel: kernel size (must be a tuple)
        - stride: moving stride
        - padding: padding size
        - dilation: stride in atrous convolution
        - group: number of groups to be divided
        - w_init: weight initializer
        - w_param: options for initializing weight
        - b_init: bias initializer
        - b_param: options for initializing bias
        - scope: name scope
        '''
        super(Conv, self).__init__(scope)
        dim = len(kernel)
        if dim == 1:
            conv_fn = layers.Conv1D
            pad_fn = layers.ZeroPadding1D
        elif dim == 2:
            conv_fn = layers.Conv2D
            pad_fn = layers.ZeroPadding2D
        elif dim == 3:
            conv_fn = layers.Conv3D
            pad_fn = layers.ZeroPadding3D
        else:
            raise Exception('NEBULAE ERROR ⨷ %d-d convolution is not supported.' % dim)

        if isinstance(stride, int):
            stride = dim * [stride]
        if isinstance(dilation, int):
            dilation = dim * [dilation]
        if isinstance(padding, int):
            padding = dim * [[padding, padding]]
        elif isinstance(padding, (list, tuple)):
            padding = [(padding[2*d], padding[2*d+1]) for d in range(dim-1, -1, -1)]

        self.pad = pad_fn(padding)
        bias = False if isinstance(b_init, Void) else True
        self.conv = conv_fn(out_chs, kernel, strides=stride, dilation_rate=dilation, groups=group,
                            use_bias=bias, kernel_initializer=w_init.iniz, bias_initializer=b_init.iniz,
                            kernel_regularizer=regularizers.l2(1), bias_regularizer=regularizers.l2(1))

    def run(self, x):
        x = self.pad(x)
        y = self.conv(x)
        return y



class _Interpad(layers.Layer):
    def __init__(self, stride: list):
        super(_Interpad, self).__init__()
        self.stride = stride

    def call(self, x):
        dim = len(x.shape)
        size = x.shape[2:]
        assert len(size) == len(self.stride)
        for d in range(dim-2):
            step = list(x.shape[0:2]) + [1 if i==d else size[i] for i in range(len(size))]
            zeros = tf.zeros_like(tf.slice(x, dim * [0], step))
            lines = tf.split(x, size[d], axis=d+2)
            padded = [lines[0]]
            for l in lines[1:]:
                padded.extend([zeros, l])
            x = tf.concat(padded, axis=d+2)
            size = x.shape[2:]

class TransConv(Craft):
    def __init__(self, in_chs, out_chs, out_size, kernel: tuple, stride=1, padding=0, dilation=1, group=1,
                 w_init=XavierNorm(), b_init=Zeros(), scope='TRANSCONV'):
        '''
        Args:
        - in_chs: input channel
        - out_chs: output channel
        - out_size: output size
        - kernel: kernel size (must be a tuple)
        - stride: moving stride
        - padding: padding size
        - dilation: stride in atrous convolution
        - group: number of groups to be divided
        - w_init: weight initializer
        - w_param: options for initializing weight
        - b_init: bias initializer
        - b_param: options for initializing bias
        - scope: name scope
        '''
        super(TransConv, self).__init__(scope)
        dim = len(kernel)
        if dim == 1:
            conv_fn = layers.Conv1D
            pad_fn = layers.ZeroPadding1D
        elif dim == 2:
            conv_fn = layers.Conv2D
            pad_fn = layers.ZeroPadding1D
        elif dim == 3:
            conv_fn = layers.Conv3D
            pad_fn = layers.ZeroPadding1D
        else:
            raise Exception('NEBULAE ERROR ⨷ %d-d convolution is not supported.' % dim)

        if isinstance(stride, int):
            stride = dim * [stride]
        if isinstance(dilation, int):
            dilation = dim * [dilation]
        if isinstance(padding, int):
            padding = dim * [2 * padding]
        elif isinstance(padding, (list, tuple)):
            padding = [padding[2*d] + padding[2*d+1] for d in range(dim-1, -1, -1)]

        in_size = []
        total_pad = []
        exter_pad = []
        for d in range(dim):
            in_size.append((out_size[d] + padding[d] - dilation[d]*(kernel[d]-1) -1) // stride[d] + 1)
            total_pad.append(dilation[d]*(kernel[d]-1) + out_size[d] - in_size[d])
            inter_pad = (in_size[d] - 1) * (stride[d] - 1)
            exter_pad.append(((total_pad[d] - inter_pad)//2, (total_pad[d] - inter_pad)-(total_pad[d] - inter_pad)//2))

        # TODO: [to be verified] interpolate stride-1 zeros between every two lines of input
        self.inter_pad = _Interpad(stride)
        self.exter_pad = pad_fn(exter_pad)
        bias = False if isinstance(b_init, Void) else True
        self.conv = conv_fn(out_chs, kernel, strides=1, padding=padding, dilation_rate=dilation, groups=group,
                            use_bias=bias, kernel_initializer=w_init.iniz, bias_initializer=b_init.iniz,
                            kernel_regularizer=regularizers.l2(1), bias_regularizer=regularizers.l2(1))

    def run(self, x):
        y = self.conv(self.exter_pad(self.inter_pad(x)))
        return y



class GraphConvAdj(Craft):
    def __init__(self, in_chs, out_chs, w_init=XavierNorm(), b_init=Zeros(), scope='GRAPHCONVADJ'):
        super(GraphConvAdj, self).__init__(scope)
        self.w = self.add_weight(
            shape=(in_chs, out_chs), initializer=w_init.iniz, trainable=True
        )
        self.has_bias = False if isinstance(b_init, Void) else True
        if self.has_bias:
            self.b = self.add_weight(shape=(out_chs,), initializer=b_init.iniz, trainable=True)

    def run(self, x, op):
        support = tf.matmul(x, self.w)
        y = tf.sparse.sparse_dense_matmul(op, support)
        if self.has_bias:
            return y + self.b
        else:
            return y



class Dense(Craft):
    def __init__(self, in_chs, out_chs, axis=-1,
                 w_init=XavierNorm(), b_init=Zeros(), scope='DENSE'):
        super(Dense, self).__init__(scope)
        if axis == 0:
            raise Exception('NEBULAE ERROR ⨷ you cannot apply dense layer along batch axis.')
        else:
            self.axis = axis

        bias = False if isinstance(b_init, Void) else True
        self.fc = layers.Dense(out_chs, use_bias=bias, kernel_initializer=w_init.iniz, bias_initializer=b_init.iniz,
                               kernel_regularizer=regularizers.l2(1), bias_regularizer=regularizers.l2(1))

    def run(self, x):
        if self.axis == -1:
            y = self.fc(x)
        else:
            dim = x.ndim()
            permuted = [i for i in range(dim)]
            permuted = permuted[:self.axis] + permuted[self.axis + 1:] + [self.axis]
            x = x.transpose(*permuted)
            y = self.fc(x)
            permuted = [i for i in range(dim)]
            permuted = permuted[:self.axis] + [dim - 1] + permuted[self.axis:-1]
            y = y.transpose(*permuted)
        return y



class Embed(Craft):
    def __init__(self, ntoken, token_dim, scope='EMBED'):
        super(Embed, self).__init__(scope)
        self.embd = layers.Embedding(ntoken, token_dim, embeddings_regularizer=regularizers.l2(1))

    def run(self, x):
        y = self.embd(x)
        return y



class Identity(Craft):
    def __init__(self, scope='IDENTITY'):
        super(Identity, self).__init__(scope)

    def run(self, x):
        return x



# ---------------------------------- Manipulation ------------------------------------ #

class EMA(object):
    def __init__(self, hull, decay_fn=lambda x: 0.9, scope='EMA'):
        self.scope = scope
        self.step = 0
        self.decay_fn = decay_fn
        self.hull = hull
        self.hull.swapped = False
        self.hull._in_ema_mode = True
        self._subcraft = []
        self._collect(self.hull)

    def _collect(self, craft):
        for k, v in craft.__dict__.items():
            if isinstance(v, Craft):
                v._in_ema_mode = True
                self._subcraft.append(v)
                self._collect(v)

    def _toggle(self):
        for c in self._subcraft:
            c.swapped = self.hull.swapped

    def __getattr__(self, item: str):
        # EMA is basically a wrapper of model i.e. self.hull,
        # hence we add the following lines to make sure the
        # attributes in the model is capable of being addressed.
        if hasattr(self, 'hull'):
            if hasattr(self.hull, item):
                return getattr(self.hull, item)
        raise AttributeError("'{}' object has no attribute '{}'".format(type(self).__name__, item))

    def swap(self):
        if tf.distribute.in_cross_replica_context():
            self._optz.swap_weights()
        else:
            if tf.keras.backend.image_data_format() == 'channels_first':
                device = 'gpu'
            else:
                device = 'cpu'
            with tf.distribute.OneDeviceStrategy(device='/%s:0' % device).scope():
                self._optz.swap_weights()
        self.hull.swapped = not self.hull.swapped
        self._toggle()

    def update(self):
        self.step += 1

    def __getitem__(self, key):
        paths = key.split('/')
        craft = self.hull
        for p in paths[:-1]:
            craft = getattr(craft, p)
        if paths[-1] == '':
            return craft
        else:
            return craft.__dict[paths[-1]]

    def __setitem__(self, key, value):
        self.hull.__dict[key] = value

    def __call__(self, *args, **kwargs):
        return self.hull(*args, **kwargs)



class Retroact(Craft):
    def __init__(self, scope='RETROACT'):
        super(Retroact, self).__init__(scope)
        if backend.image_data_format() == 'channels_first':
            self.paxis = [2, 3]
            self.caxis = 1
        else:
            self.paxis = [1, 2]
            self.caxis = -1

    def run(self, i, o, pb):
        i_grad = pb.gradient(o, i)

        pooled_grad = tf.reduce_mean(i_grad, axis=self.paxis, keepdims=True)

        i *= pooled_grad

        hmap = tf.reduce_mean(i, axis=self.caxis, keepdims=True) # average over channels #1
        hmap = tf.where(hmap<0, tf.zeros_like(hmap), hmap) # pull negative values up to zero
        hmap /= tf.reduce_max(hmap, axis=[1, 2, 3], keepdims=True) + 1e-6

        del i_grad

        return hmap



# ----------------------------------- Mathmatic -------------------------------------- #

class Sqrt(Craft):
    def __init__(self, scope='SQRT'):
        super(Sqrt, self).__init__(scope)

    def run(self, t):
        y = tf.sqrt(t)
        return y



# ------------------------------------ Polyadic -------------------------------------- #

class Concat(Craft):
    def __init__(self, scope='CONCAT'):
        super(Concat, self).__init__(scope)

    def run(self, t, axis=-1):
        y = tf.concat(t, axis)
        return y



# ------------------------------------- Sparse --------------------------------------- #

class D2S(Craft):
    def __init__(self, scope='D2S'):
        super(D2S, self).__init__(scope)

    def run(self, x):
        y = tf.sparse.from_dense(x)
        return y



class S2D(Craft):
    def __init__(self, scope='S2D'):
        super(S2D, self).__init__(scope)

    def run(self, x):
        y = tf.sparse.to_dense(x)
        return y



class SpDot(Craft):
    def __init__(self, scope='SPDOT'):
        super(SpDot, self).__init__(scope)

    def run(self, x, y, in_batch=False):
        z = tf.sparse.sparse_dense_matmul(x, y)
        return z



# ----------------------------------- Statistics ------------------------------------- #

class Clip(Craft):
    def __init__(self, intrinsic=False, scope='CLIP'):
        super(Clip, self).__init__(scope)
        if intrinsic:
            self.clipper = tfp.math.clip_by_value_preserve_gradient
        else:
            self.clipper = tf.clip_by_value

    def run(self, x, ranges):
        if isinstance(ranges, tuple):
            assert len(ranges)==2
        else:
            ranges = (-ranges, ranges)
        return self.clipper(x, *ranges)



class Mean(Craft):
    def __init__(self, scope='MEAN'):
        super(Mean, self).__init__(scope)

    def run(self, x, axis=None):
        if axis is None:
            y = tf.reduce_mean(x)
        else:
            y = tf.reduce_mean(x, axis=axis)
        return y



class Max(Craft):
    def __init__(self, scope='MAX'):
        super(Max, self).__init__(scope)

    def run(self, x, axis=None):
        if axis is None:
            y = tf.reduce_max(x)
        else:
            y = tf.reduce_max(x, axis=axis)
        return y



class Min(Craft):
    def __init__(self, scope='MIN'):
        super(Min, self).__init__(scope)

    def run(self, x, axis=None):
        if axis is None:
            y = tf.reduce_min(x)
        else:
            y = tf.reduce_min(x, axis=axis)
        return y



class Sum(Craft):
    def __init__(self, scope='SUM'):
        super(Sum, self).__init__(scope)

    def run(self, x, axis=None):
        if axis is None:
            y = tf.reduce_sum(x)
        else:
            y = tf.reduce_sum(x, axis=axis)
        return y



class Dot(Craft):
    def __init__(self, scope='DOT'):
        super(Dot, self).__init__(scope)

    def run(self, x, y, in_batch=False):
        z = x @ y
        return z



class Grad(Craft):
    def __init__(self, scope='GRAD'):
        super(Grad, self).__init__(scope)

    def run(self, i, o, pb):
        g = pb.gradient(o, i)
        return g



# ---------------------------------- Linear Algebra ---------------------------------- #

class QR(Craft):
    def __init__(self, scope='QR'):
        super(QR, self).__init__(scope)

    def run(self, m, reduced=False):
        q, r = tf.linalg.qr(m, full_matrices=not reduced)
        return q, r



class Eigen(Craft):
    def __init__(self, scope='EIGEN'):
        super(Eigen, self).__init__(scope)

    def run(self, m, in_pair=True):
        if in_pair:
            return tf.linalg.eig(m)
        else:
            return tf.linalg.eigvals(m)



class MaxEigen(Craft):
    def __init__(self, niters=3, scope='MAXEIGEN'):
        super(MaxEigen, self).__init__(scope)
        self.niters = niters

    def run(self, m, in_pair=True):
        assert (m==tf.transpose(m)).all(), 'NEBULAE ERROR ⨷ only when m is real symmetric matrix can MaxEigen be applied.'
        x = tf.random.normal((m.shape[-1], 1), 0, 1)
        for _ in range(self.niters):
            x = m @ x
            x = tf.math.l2_normalize(x)
        eigval_max = tf.transpose(x) @ m @ x
        if not in_pair:
            return eigval_max
        else:
            eigvec_max = (m @ x) / eigval_max
            return eigval_max, eigvec_max



# ------------------------------------- Resizer -------------------------------------- #

class Reshape(Craft):
    def __init__(self, scope='RESHAPE'):
        super(Reshape, self).__init__(scope)

    def run(self, x, shape):
        y = tf.reshape(x, shape)
        return y



class Permute(Craft):
    def __init__(self, scope='PERMUTE'):
        super(Permute, self).__init__(scope)

    def run(self, x, order):
        y = tf.transpose(x, order)
        return y



class Upscale(Craft):
    def __init__(self, scale: tuple, interp=NEAREST, scope='UPS'):
        super(Upscale, self).__init__(scope)
        dim = len(scale)
        if dim == 1:
            self.fn = layers.UpSampling1D(scale_factor=scale, mode=TF_INTERP[interp])
        elif dim == 2:
            self.fn = layers.UpSampling2D(scale_factor=scale, mode=TF_INTERP[interp])
        elif dim == 3:
            self.fn = layers.UpSampling3D(scale_factor=scale, mode=TF_INTERP[interp])
        else:
            raise Exception('NEBULAE ERROR ⨷ %d-d upscaling is not supported.' % dim)

    def run(self, x):
        y = self.fn(x)
        return y



class _MP(layers.Layer):
    def __init__(self, dim):
        super(_MP, self).__init__()
        if dim == 1:
            self.pool = layers.GlobalMaxPool1D()
        elif dim == 2:
            self.pool = layers.GlobalMaxPool2D()
        elif dim == 3:
            self.pool = layers.GlobalMaxPool3D()

    def call(self, x):
        if backend.image_data_format() == 'channels_first':
            return tf.expand_dims(tf.expand_dims(self.pool(x), axis=-1), axis=-1)
        else:
            return tf.expand_dims(tf.expand_dims(self.pool(x), axis=-2), axis=-2)

class MaxPool(Craft):
    def __init__(self, kernel: tuple, stride=2, padding=0, scope='MPOOL'):
        super(MaxPool, self).__init__(scope)
        dim = len(kernel)
        is_global = True if kernel[-1] < 0 else False
        if is_global:
            assert padding == 0

        if dim == 1:
            if is_global:
                self.pool = _MP(1)
            else:
                self.pool = layers.MaxPool1D(kernel, stride)
            pad_fn = layers.ZeroPadding1D
        elif dim == 2:
            if is_global:
                self.pool = _MP(2)
            else:
                self.pool = layers.MaxPool2D(kernel, stride)
            pad_fn = layers.ZeroPadding2D
        elif dim == 3:
            if is_global:
                self.pool = _MP(3)
            else:
                self.pool = layers.MaxPool3D(kernel, stride)
            pad_fn = layers.ZeroPadding3D
        else:
            raise Exception('NEBULAE ERROR ⨷ %d-d pooling is not supported.' % dim)

        if isinstance(padding, int):
            padding = dim * [[padding, padding]]
        elif isinstance(padding, (list, tuple)):
            padding = [(padding[2*d], padding[2*d+1]) for d in range(dim-1, -1, -1)]

        self.pad = pad_fn(padding)

    def run(self, x):
        y = self.pool(self.pad(x))
        return y



class _AP(layers.Layer):
    def __init__(self, dim):
        super(_AP, self).__init__()
        if dim == 1:
            self.pool = layers.GlobalAveragePooling1D()
        elif dim == 2:
            self.pool = layers.GlobalAveragePooling2D()
        elif dim == 3:
            self.pool = layers.GlobalAveragePooling3D()

    def call(self, x):
        if backend.image_data_format() == 'channels_first':
            return tf.expand_dims(tf.expand_dims(self.pool(x), axis=-1), axis=-1)
        else:
            return tf.expand_dims(tf.expand_dims(self.pool(x), axis=-2), axis=-2)

class AvgPool(Craft):
    def __init__(self, kernel: tuple, stride=2, padding=0, scope='APOOL'):
        super(AvgPool, self).__init__(scope)
        dim = len(kernel)
        is_global = True if kernel[-1]<0 else False
        if is_global:
            assert padding == 0

        if dim == 1:
            if is_global:
                self.pool = _AP(1)
            else:
                self.pool = layers.AveragePooling1D(kernel, stride)
            pad_fn = layers.ZeroPadding1D
        elif dim == 2:
            if is_global:
                self.pool = _AP(2)
            else:
                self.pool = layers.AveragePooling2D(kernel, stride)
            pad_fn = layers.ZeroPadding2D
        elif dim == 3:
            if is_global:
                self.pool = _AP(3)
            else:
                self.pool = layers.AveragePooling3D(kernel, stride)
            pad_fn = layers.ZeroPadding3D
        else:
            raise Exception('NEBULAE ERROR ⨷ %d-d pooling is not supported.' % dim)

        if isinstance(padding, int):
            padding = dim * [[padding, padding]]
        elif isinstance(padding, (list, tuple)):
            padding = [(padding[2*d], padding[2*d+1]) for d in range(dim-1, -1, -1)]

        self.pad = pad_fn(padding)

    def run(self, x):
        y = self.pool(self.pad(x))
        return y



# ------------------------------------ Activation ------------------------------------ #

class Relu(Craft):
    def __init__(self, scope='RELU'):
        super(Relu, self).__init__(scope)
        self.actv = layers.ReLU()

    def run(self, x):
        y = self.actv(x)
        return y



class LRelu(Craft):
    def __init__(self, alpha=0.2, scope='LRELU'):
        super(LRelu, self).__init__(scope)
        self.actv = layers.LeakyReLU(alpha)

    def run(self, x):
        y = self.actv(x)
        return y



class Tanh(Craft):
    def __init__(self, scope='TANH'):
        super(Tanh, self).__init__(scope)

    def run(self, x):
        y = tf.math.tanh(x)
        return y



class Sigm(Craft):
    def __init__(self, scope='SIGM'):
        super(Sigm, self).__init__(scope)

    def run(self, x):
        y = tf.math.sigmoid(x)
        return y



class Sftm(Craft):
    def __init__(self, axis=-1, scope='SFTM'):
        super(Sftm, self).__init__(scope)
        self.actv = layers.Softmax(axis)

    def run(self, x):
        y = self.actv(x)
        return y



class Sftp(Craft):
    def __init__(self, scope='SFTP'):
        super(Sftp, self).__init__(scope)

    def run(self, x):
        y = tf.math.softplus(x)
        return y



# ------------------------------------ Distributing ------------------------------------ #

class Dropout(Craft):
    def __init__(self, p_drop, dim, scope='DROPOUT'):
        super(Dropout, self).__init__(scope)
        self.dp = layers.Dropout(p_drop)

    def run(self, x):
        y = self.dp(x)
        return y



class BN(Craft):
    def __init__(self, out_chs, dim, mmnt=0.9, resilient=True, scope='BN'):
        super(BN, self).__init__(scope)
        self.norm = layers.experimental.SyncBatchNormalization(1, momentum=mmnt, center=resilient, scale=resilient, epsilon=1e-5)

    def run(self, x):
        y = self.norm(x)
        return y


'''
class CBN(Craft):
    def __init__(self, in_chs, out_chs, dim, mmnt=0.9, scope='CBN'):
        super(CBN, self).__init__(scope)
        if dim == 1:
            norm_fn = nn.BatchNorm1d
        elif dim == 2:
            norm_fn = nn.BatchNorm2d
        elif dim == 3:
            norm_fn = nn.BatchNorm3d
        else:
            raise Exception('NEBULAE ERROR ⨷ %d-d CN is not supported.' % dim)
        self.norm = norm_fn(out_chs, momentum=1 - mmnt, affine=False, eps=1e-5)
        self.relu = nn.ReLU()
        self.gamma_1 = nn.Linear(in_chs, in_chs // 2)
        self.gamma_2 = nn.Linear(in_chs // 2, out_chs)
        self.beta_1 = nn.Linear(in_chs, in_chs // 2)
        self.beta_2 = nn.Linear(in_chs // 2, out_chs)

    def run(self, x, z):
        y = self.norm(x)

        g = self.gamma_1(z)
        g = self.relu(g)
        g = self.gamma_2(g)

        b = self.beta_1(z)
        b = self.relu(b)
        b = self.beta_2(b)

        for _ in range(x.ndim - 2):
            g = g.unsqueeze(-1)
            b = b.unsqueeze(-1)

        self.weight = g
        self.bias = b
        y = self.weight * y + self.bias

        return y



class IN(Craft):
    def __init__(self, out_chs, dim, mmnt=0.9, resilient=True, scope='IN'):
        super(IN, self).__init__(scope)
        if dim == 1:
            norm_fn = nn.InstanceNorm1d
        elif dim == 2:
            norm_fn = nn.InstanceNorm2d
        elif dim == 3:
            norm_fn = nn.InstanceNorm3d
        else:
            raise Exception('NEBULAE ERROR ⨷ %d-d IN is not supported.' % dim)
        self.norm = norm_fn(out_chs, momentum=1 - mmnt, affine=resilient, eps=1e-5)

    def run(self, x):
        y = self.norm(x)
        return y



class CIN(Craft):
    def __init__(self, in_chs, out_chs, dim, mmnt=0.9, scope='CIN'):
        super(CIN, self).__init__(scope)
        if dim == 1:
            norm_fn = nn.InstanceNorm1d
        elif dim == 2:
            norm_fn = nn.InstanceNorm2d
        elif dim == 3:
            norm_fn = nn.InstanceNorm3d
        else:
            raise Exception('NEBULAE ERROR ⨷ %d-d CN is not supported.' % dim)
        self.norm = norm_fn(out_chs, momentum=1 - mmnt, affine=False, eps=1e-5)
        self.relu = nn.ReLU()
        self.gamma_1 = nn.Linear(in_chs, in_chs // 2)
        self.gamma_2 = nn.Linear(in_chs // 2, out_chs)
        self.beta_1 = nn.Linear(in_chs, in_chs // 2)
        self.beta_2 = nn.Linear(in_chs // 2, out_chs)

    def run(self, x, z):
        y = self.norm(x)

        g = self.gamma_1(z)
        g = self.relu(g)
        g = self.gamma_2(g)

        b = self.beta_1(z)
        b = self.relu(b)
        b = self.beta_2(b)

        for _ in range(x.ndim - 2):
            g = g.unsqueeze(-1)
            b = b.unsqueeze(-1)

        self.weight = g
        self.bias = b
        y = self.weight * y + self.bias

        return y



class LN(Craft):
    def __init__(self, norm_shape, resilient=True, scope='LN'):
        super(LN, self).__init__(scope)
        norm_shape = tuple([norm_shape[-1]] + [ns for ns in norm_shape[:-1]])
        self.norm = nn.LayerNorm(norm_shape, elementwise_affine=resilient, eps=1e-5)

    def run(self, x):
        y = self.norm(x)
        return y



class SN(Craft):
    def __init__(self, craft, niter=3, eps=1e-12, scope='SN'):
        super(SN, self).__init__(scope)
        self.name = 'weight'
        if isinstance(craft, (Conv, TransConv)):
            self.key = 'conv/'
        elif isinstance(craft, Dense):
            self.key = 'fc/'
        elif isinstance(craft, Embed):
            self.key = 'embd/'
        elif isinstance(craft, (BN, IN, LN)):
            self.key = 'norm/'
        elif isinstance(craft, (CBN, CIN)):
            self.key = ''
        else:
            raise Exception('NEBULAE ERROR ⨷ SN does not support %s layer.' % type(craft))
        self.craft = craft
        self.hull = craft[self.key]
        self.niter = niter
        self.eps = eps
        if not self._made_params():
            self._make_params()

    def l2normalize(self, v):
        return v / (v.norm() + self.eps)

    def _update_u_v(self):
        if not self._made_params():
            self._make_params()
        w = getattr(self.hull, self.name)
        u = getattr(self.hull, self.name + "_u")

        height = w.data.shape[0]
        for _ in range(self.niter):
            v = self.l2normalize(torch.mv(torch.t(w.view(height, -1).data), u))
            u = self.l2normalize(torch.mv(w.view(height, -1).data, v))

        setattr(self.hull, self.name + "_u", u)
        w.data = w.data / torch.dot(u, torch.mv(w.view(height, -1).data, v))

    def _made_params(self):
        return hasattr(self.hull, self.name + "_u")

    def _make_params(self):
        w = getattr(self.hull, self.name)

        height = w.data.shape[0]
        width = w.view(height, -1).data.shape[1]

        u = self.l2normalize(w.data.new(height).normal_(0, 1))

        self.hull.register_buffer(self.name + "_u", u)

    def run(self, *args, **kwargs):
        self._update_u_v()
        return self.craft.run(*args, **kwargs)


'''
# ------------------------------------ Loss ------------------------------------ #

class MAE(Craft):
    def __init__(self, scope='MAE'):
        super(MAE, self).__init__(scope)
        self.cost = losses.MeanAbsoluteError(reduction=losses.Reduction.SUM)

    def run(self, x, y):
        z = self.cost(x, y)
        return z/x.shape[0]



class MSE(Craft):
    def __init__(self, scope='MSE'):
        super(MSE, self).__init__(scope)
        self.cost = losses.MeanSquaredError(reduction=losses.Reduction.SUM)

    def run(self, x, y):
        z = self.cost(x, y)
        return z/x.shape[0]



class SigmXE(Craft):
    def __init__(self, scope='SIGMXE'):
        super(SigmXE, self).__init__(scope)
        self.cost = losses.BinaryCrossentropy(from_logits=True, reduction=losses.Reduction.SUM)

    def run(self, x, y):
        z = self.cost(y, x)
        return z/x.shape[0]



class SftmXE(Craft):
    def __init__(self, is_one_hot, scope='SFTMXE'):
        super(SftmXE, self).__init__(scope)
        if is_one_hot:
            self.cost = losses.CategoricalCrossentropy(from_logits=True, reduction=losses.Reduction.SUM)
        else:
            self.cost = losses.SparseCategoricalCrossentropy(from_logits=True, reduction=losses.Reduction.SUM)

    def run(self, x, y):
        z = self.cost(y, x)
        return z/x.shape[0]



# ------------------------------------ Metric ------------------------------------ #

class AccCls(Craft):
    def __init__(self, multi_class, is_one_hot, scope='ACCCLS'):
        super(AccCls, self).__init__(scope)
        if multi_class:
            assert not is_one_hot
        self.mulcls = multi_class
        self.ioh = is_one_hot

    def run(self, x, y):
        data_type = x.dtype
        if self.mulcls: # include binary classification as well
            x = tf.round(x)
            correct = tf.reduce_mean(tf.cast(x == y, dtype=data_type), axis=-1)
            z = tf.reduce_mean(tf.cast(correct == 1, dtype=data_type))
        else:
            if self.ioh:
                y = tf.argmax(y, axis=-1)
            x = tf.argmax(x, axis=-1)
            z = tf.reduce_mean(tf.cast(x == y, dtype=data_type))
        return z


# ------------------------------------ Optimizer ------------------------------------ #

class WarmUpWrapper(tf.keras.optimizers.schedules.LearningRateSchedule):
    def __init__(self, warmup, scheduler):
        self.warmup = warmup

        def retnum(s):
            return self.lr

        if isinstance(scheduler, (int, float)):
            self.lr = tf.cast(scheduler, tf.float32)
            self.scheduler = retnum
        else:
            self.lr = scheduler.get_config()['initial_learning_rate']
            self.scheduler = scheduler

    def __call__(self, step):
        def warmlr(s):
            return self.lr * tf.cast(s, tf.float32) / tf.cast(self.warmup, tf.float32)

        return tf.cond(step<self.warmup, lambda:warmlr(step), lambda:self.scheduler(step-self.warmup))

    def get_config(self):
        return self.scheduler.get_config()



class StepLR(object):
    def __init__(self, period, factor):
        self.period = period
        self.factor = factor

    def __call__(self, lr):
        return optimizers.schedules.ExponentialDecay(lr, self.period, self.factor, staircase=True)



class PolyLR(object):
    def __init__(self, cutoff, power):
        self.cutoff = cutoff
        self.power = power

    def __call__(self, lr):
        return optimizers.schedules.PolynomialDecay(lr, self.cutoff, power=self.power)



class CosLR(object):
    def __init__(self, period):
        self.period = period

    def __call__(self, lr):
        return optimizers.schedules.CosineDecay(lr, self.period, lr * 0.001)



class ExpLR(object):
    def __init__(self, period, factor):
        self.period = period
        self.factor = factor

    def __call__(self, lr):
        return optimizers.schedules.ExponentialDecay(lr, self.period, self.factor)



class WavyLR(object):
    def __init__(self, period, dampen):
        self.period = period
        self.dampen = dampen

    def __call__(self, lr):
        return optimizers.schedules.CosineDecayRestarts(lr, self.period // 2, t_mul=1, m_mul=self.dampen)



class Momentum(Craft):
    def __init__(self, hull, lr, mmnt=0.9, wd=0, lr_decay=None, warmup=0,
                 grad_limit=-1, grad_accum=1, update_scope=None, scope='MOMENTUM'):
        super(Momentum, self).__init__(scope)
        self.hull = hull
        self.wd = wd / 2.

        self.update_scope = update_scope
        self.grad_accum = grad_accum
        self.cnt = 0
        if grad_limit<0:
            grad_limit = None
        if lr_decay is not None:
            lr = lr_decay(lr)
        if warmup>0:
            lr = WarmUpWrapper(warmup, lr)
        self.optz = optimizers.SGD(lr, momentum=mmnt, clipvalue=grad_limit)
        if getattr(self.hull, '_in_ema_mode', False):
            self.optz = tfa.optimizers.MovingAverage(self.optz)

            def average_op(itself, var, average_var):
                decay = tf.constant(self.hull.decay_fn(self.hull.step), dtype=tf.float32)
                return tf.keras.backend.moving_average_update(average_var, var, decay)

            self.optz.average_op = MethodType(average_op, self.optz)
            self.hull._optz = self.optz

    def _getVar(self):
        # select parameters await updating
        if self.update_scope is None:
            self.update_var = self.hull.trainable_weights
        else:
            if isinstance(self.update_scope, str):
                self.update_scope = [self.update_scope]
            self.update_var = []
            for us in self.update_scope:
                paths = us.split('/')
                craft = self.hull
                for p in paths:
                    craft = getattr(craft, p)
                self.update_var.extend(craft.trainable_weights)

    def run(self, target):
        # variables have not been created before runtime
        if not hasattr(self, 'update_var'):
            self._getVar()
        target += tf.add_n(self.hull.losses)
        curr_grads = self.hull._rudder.gradient(target, self.update_var)
        curr_grads = [g if g is not None else tf.zeros_like(v) for g, v in zip(curr_grads, self.update_var)]
        if self.cnt == 0:
            self.grads = curr_grads
        else:
            self.grads = [self.grads[i] + curr_grads[i] for i in len(curr_grads)]
        self.cnt += 1
        if self.cnt == self.grad_accum:
            # if self.e>0:
            #     import pdb;
            #     pdb.set_trace()
            self.cnt = 0
            self.optz.apply_gradients(zip(self.grads, self.update_var))



class Nesterov(Craft):
    def __init__(self, hull, lr, mmnt=0.9, wd=0, lr_decay=None, warmup=0,
                 grad_limit=-1, grad_accum=1, update_scope=None, scope='NESTEROV'):
        super(Nesterov, self).__init__(scope)
        self.hull = hull
        self.wd = wd / 2.

        self.update_scope = update_scope
        self.grad_accum = grad_accum
        self.cnt = 0
        if grad_limit < 0:
            grad_limit = None
        if lr_decay is not None:
            lr = lr_decay(lr)
        if warmup>0:
            lr = WarmUpWrapper(warmup, lr)
        self.optz = optimizers.SGD(lr, momentum=mmnt, nesterov=True, clipvalue=grad_limit)
        if getattr(self.hull, '_in_ema_mode', False):
            self.optz = tfa.optimizers.MovingAverage(self.optz)

            def average_op(itself, var, average_var):
                decay = tf.constant(self.hull.decay_fn(self.hull.step), dtype=tf.float32)
                return tf.keras.backend.moving_average_update(average_var, var, decay)

            self.optz.average_op = MethodType(average_op, self.optz)
            self.hull._optz = self.optz

    def _getVar(self):
        # select parameters await updating
        if self.update_scope is None:
            self.update_var = self.hull.trainable_weights
        else:
            if isinstance(self.update_scope, str):
                self.update_scope = [self.update_scope]
            self.update_var = []
            for us in self.update_scope:
                paths = us.split('/')
                craft = self.hull
                for p in paths:
                    craft = getattr(craft, p)
                self.update_var.extend(craft.trainable_weights)

    def run(self, target):
        # variables have not been created before runtime
        if not hasattr(self, 'update_var'):
            self._getVar()
        target += tf.add_n(self.hull.losses)
        curr_grads = self.hull._rudder.gradient(target + self.wd*tf.add_n(self.hull.losses), self.update_var)
        # curr_grads = [g if g is not None else tf.zeros_like(v) for g, v in zip(curr_grads, self.update_var)]
        if self.cnt == 0:
            self.grads = curr_grads
        else:
            self.grads = [self.grads[i] + curr_grads[i] for i in len(curr_grads)]
        self.cnt += 1
        if self.cnt == self.grad_accum:
            self.cnt = 0
            self.optz.apply_gradients(zip(self.grads, self.update_var))



class RMSProp(Craft):
    def __init__(self, hull, lr, mmnt=0., wd=0, lr_decay=None, warmup=0,
                 grad_limit=-1, grad_accum=1, update_scope=None, scope='RMSPROP'):
        super(RMSProp, self).__init__(scope)
        self.hull = hull
        self.wd = wd / 2.

        self.update_scope = update_scope
        self.grad_accum = grad_accum
        self.cnt = 0
        if grad_limit < 0:
            grad_limit = None
        if lr_decay is not None:
            lr = lr_decay(lr)
        if warmup>0:
            lr = WarmUpWrapper(warmup, lr)
        self.optz = optimizers.RMSprop(lr, momentum=mmnt, clipvalue=grad_limit)
        if getattr(self.hull, '_in_ema_mode', False):
            self.optz = tfa.optimizers.MovingAverage(self.optz)

            def average_op(itself, var, average_var):
                decay = tf.constant(self.hull.decay_fn(self.hull.step), dtype=tf.float32)
                return tf.keras.backend.moving_average_update(average_var, var, decay)

            self.optz.average_op = MethodType(average_op, self.optz)
            self.hull._optz = self.optz

    def _getVar(self):
        # select parameters await updating
        if self.update_scope is None:
            self.update_var = self.hull.trainable_weights
        else:
            if isinstance(self.update_scope, str):
                self.update_scope = [self.update_scope]
            self.update_var = []
            for us in self.update_scope:
                paths = us.split('/')
                craft = self.hull
                for p in paths:
                    craft = getattr(craft, p)
                self.update_var.extend(craft.trainable_weights)

    def run(self, target):
        # variables have not been created before runtime
        if not hasattr(self, 'update_var'):
            self._getVar()
        target += tf.add_n(self.hull.losses)
        curr_grads = self.hull._rudder.gradient(target + self.wd*tf.add_n(self.hull.losses), self.update_var)
        # curr_grads = [g if g is not None else tf.zeros_like(v) for g, v in zip(curr_grads, self.update_var)]
        if self.cnt == 0:
            self.grads = curr_grads
        else:
            self.grads = [self.grads[i] + curr_grads[i] for i in len(curr_grads)]
        self.cnt += 1
        if self.cnt == self.grad_accum:
            self.cnt = 0
            self.optz.apply_gradients(zip(self.grads, self.update_var))



class Adam(Craft):
    def __init__(self, hull, lr, mmnt1=0.9, mmnt2=0.999, wd=0, lr_decay=None, warmup=0,
                 grad_limit=-1, grad_accum=1, update_scope=None, scope='ADAM'):
        super(Adam, self).__init__(scope)
        self.hull = hull
        self.wd = wd / 2.

        self.update_scope = update_scope
        self.grad_accum = grad_accum
        self.cnt = 0
        if grad_limit < 0:
            grad_limit = None
        if lr_decay is not None:
            lr = lr_decay(lr)
        if warmup>0:
            lr = WarmUpWrapper(warmup, lr)
        self.optz = optimizers.Adam(lr, beta_1=mmnt1, beta_2=mmnt2, clipvalue=grad_limit)
        if getattr(self.hull, '_in_ema_mode', False):
            self.optz = tfa.optimizers.MovingAverage(self.optz)

            def average_op(itself, var, average_var):
                decay = tf.constant(self.hull.decay_fn(self.hull.step), dtype=tf.float32)
                return backend.moving_average_update(average_var, var, decay)

            self.optz.average_op = MethodType(average_op, self.optz)
            self.hull._optz = self.optz

    def _getVar(self):
        # select parameters await updating
        if self.update_scope is None:
            self.update_var = self.hull.trainable_weights
        else:
            if isinstance(self.update_scope, str):
                self.update_scope = [self.update_scope]
            self.update_var = []
            for us in self.update_scope:
                paths = us.split('/')
                craft = self.hull
                for p in paths:
                    craft = getattr(craft, p)
                self.update_var.extend(craft.trainable_weights)

    def run(self, target):
        # variables have not been created before runtime
        if not hasattr(self, 'update_var'):
            self._getVar()
        target += tf.add_n(self.hull.losses)
        curr_grads = self.hull._rudder.gradient(target + self.wd*tf.add_n(self.hull.losses), self.update_var)
        # curr_grads = [g if g is not None else tf.zeros_like(v) for g, v in zip(curr_grads, self.update_var)]
        if self.cnt == 0:
            self.grads = curr_grads
        else:
            self.grads = [self.grads[i] + curr_grads[i] for i in len(curr_grads)]
        self.cnt += 1
        if self.cnt == self.grad_accum:
            self.cnt = 0
            self.optz.apply_gradients(zip(self.grads, self.update_var))


#!/usr/bin/env python
'''
Created by Seria at 02/11/2018 2:50 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-

from .decorator import Timer
from .utility import hotvec2mtx, mtx2hotvec, den2spa, spa2den, randTruncNorm, parseConfig, recordConfig, byte2arr, \
                    joinImg, plotInOne, ver2num, GPUtil

import os
core = os.environ.get('NEB_CORE', 'PYTORCH')
if core.upper() == 'TENSORFLOW':
    import tensorflow as tf
    import random
    import numpy as np
    def destine(seed):
        random.seed(seed)
        np.random.seed(seed)
        tf.random.set_seed(seed)
elif core.upper() == 'PYTORCH':
    import torch
    import random
    import numpy as np
    def destine(seed):
        random.seed(seed)
        np.random.seed(seed)
        torch.manual_seed(seed)
        # if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
else:
    raise ValueError('NEBULAE ERROR ⨷ %s is an unsupported core.' % core)


#!/usr/bin/env python
'''
Created by Seria at 02/11/2018 3:38 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========

'''
# -*- coding:utf-8 -*-
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '1'
import h5py
from collections import abc
from math import ceil
import tensorflow as tf
# import tensorflow_io as tfio


class Tank(object):
    def __init__(self, data_path, data_specf, batch_size, shuffle=True, in_same_size=True, nworkers=-1,
                 fetch_fn=None, prep_fn=None, collate_fn=None):
        name = os.path.basename(data_path).split('.')[0]
        self.rank = int(os.environ.get('RANK', -1))
        # nworld = int(os.environ.get('WORLD_SIZE', 1))

        ret_struct = -1
        with h5py.File(data_path, 'r') as f:
            length = len(f[list(data_specf.keys())[0]])
            elem = fetch_fn(f, 0)
            etype = type(elem)
            # ndarray
            if etype.__module__ == 'numpy' and etype.__name__ != 'str_' and etype.__name__ != 'string_':
                ret_struct = 0
                assert len(data_specf) == 1
                for k, v in data_specf.items():
                    if v.startswith('v'):
                        v = 'string'
                    dtype = getattr(tf.dtypes, v)
                    # dshape = elem.shape
            # single value
            elif isinstance(elem, (int, float, str, bytes)):
                ret_struct = 1
                assert len(data_specf) == 1
                for k, v in data_specf.items():
                    if v.startswith('v'):
                        v = 'string'
                    dtype = getattr(tf.dtypes, v)
                    # dshape = ()
            # mapping type
            elif isinstance(elem, abc.Mapping):
                ret_struct = 2
                dtype = {}
                dshape = {}
                for k, v in data_specf.items():
                    if v.startswith('v'):
                        v = 'string'
                    dtype[k] = getattr(tf.dtypes, v)
                    # dshape[k] = elem[k].shape
            # sequential type
            elif isinstance(elem, abc.Sequence):
                ret_struct = 3
                dtype = []
                dshape = []
                for i, v in enumerate(data_specf.values()):
                    if v.startswith('v'):
                        v = 'string'
                    dtype.append(getattr(tf.dtypes, v))
                    # dshape.append(elem[i].shape)
            else:
                raise Exception('NEBULAE ERROR ⨷ %s is not a valid type of data.' % type(elem))

        class TData(tf.data.Dataset):
            @staticmethod
            def _generator():
                # Opening the file
                hdf5 = h5py.File(data_path, 'r')
                length = len(hdf5[list(data_specf.keys())[0]])

                for idx in range(length):
                    # Reading data from the file
                    ret = fetch_fn(hdf5, idx)
                    for i, (k, v) in enumerate(data_specf.items()):
                        if v.startswith('v'):
                            if ret_struct <= 1:
                                ret = ret.tobytes()
                            elif ret_struct == 2:
                                ret[k] = ret[k].tobytes()
                            else:
                                ret[i] = ret[i].tobytes()
                    yield ret

            def __new__(cls):
                dset = tf.data.Dataset.from_generator(cls._generator, output_types=dtype)#, output_shapes=dshape)
                    # output_signature=[tf.TensorSpec(shape=(), dtype=tf.int32)])
                dset.length = length
                if self.rank<=0:
                    print('+' + (49 * '-') + '+')
                    print('| \033[1;35m%-20s\033[0m fuel tank has been mounted |' % name)
                    print('+' + (49 * '-') + '+')
                return dset

        self.name = name
        self.counter = 0

        self.tdata = TData()
        length = self.tdata.length
        if prep_fn is not None:
            if nworkers<=0:
                self.tdata = self.tdata.map(prep_fn, num_parallel_calls=tf.data.experimental.AUTOTUNE)
            else:
                self.tdata = self.tdata.map(prep_fn, num_parallel_calls=nworkers)
        if shuffle:
            self.tdata = self.tdata.shuffle(8 * batch_size, reshuffle_each_iteration=True)
        self.tdata = self.tdata.batch(batch_size, drop_remainder=in_same_size)
        self.batch_size = batch_size
        if in_same_size:
            self.MPE = length // batch_size
        else:
            self.MPE = ceil(length / batch_size)
        self.tdata = self.tdata.cache().repeat().prefetch(tf.data.experimental.AUTOTUNE)

    def __del__(self):
        if self.rank <= 0:
            print('+' + (53 * '-') + '+')
            print('| \033[1;35m%-20s\033[0m fuel tank is no longer mounted |' % self.name)
            print('+' + (53 * '-') + '+')

    def __len__(self):
        return self.tdata.length

    def next(self):
        if self.counter == 0: # create a new iterator at the begining of an epoch
            self.iterator = self.tdata.__iter__()
        self.counter += 1
        if self.counter == self.MPE:
            self.counter = 0
        return self.iterator.__next__()


#!/usr/bin/env python
'''
engine_mx
Created by Seria at 12/02/2019 3:45 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
import os
import torch
from ..toolkit.utility import GPUtil

class EnginePT(object):
    '''
    Param:
    device: 'gpu' or 'cpu'
    available_gpus
    gpu_mem_fraction
    if_conserve
    least_mem
    '''
    def __init__(self, param):
        self.param = param
        # look for available gpu devices
        if self.param['device'].lower() == 'gpu':
            if len(self.param['avail_gpus']) == 0:
                gputil = GPUtil()
                gpus = gputil.available(self.param['ngpus'], self.param['least_mem'])
            else:
                gpus = self.param['avail_gpus']
            if len(gpus) < self.param['ngpus']:
                raise Exception('NEBULAE ERROR ⨷ no enough available gpu.')
            # convert gpu list to string
            str_gpus = ','.join([str(g[0]) for g in gpus])
            # set environment variable
            os.environ['CUDA_VISIBLE_DEVICES'] = str_gpus
            self.device = [torch.device('cuda:%d'%i) for i in range(self.param['ngpus'])]
            # setup multi-gpu environment
            if self.param['rank']>=0:
                torch.backends.cudnn.benchmark = True
                torch.cuda.set_device(self.param['rank'])
                torch.distributed.init_process_group(backend='nccl', init_method='env://')
            if self.param['rank']<=0:
                print('+' + 20 * '-' + '+')
                print('| Reside in Devices: |')
                print('+' + 70 * '-' + '+')
                for g in gpus:
                    print('| \033[1;36mGPU {:<2d}\033[0m | {:<25s} | {:>5d} MiB free out of {:<5d} MiB |'.format(
                        g[0], g[1], g[3], g[2]
                    ))
                    print('+' + 70 * '-' + '+')
        elif self.param['device'].lower() == 'cpu':
            assert self.param['rank']<0
            self.device = ['cpu']
            print('+' + (24 * '-') + '+')
            print('| Reside in Devices: \033[1;36mCPU\033[0m |')
            print('+' + (24 * '-') + '+')
        else:
            raise KeyError('NEBULAE ERROR ⨷ given device should be either cpu or gpu.')

        ###################################
        ##          IT WORKS !           ##
        ###################################
        from ..astrobase import dock
        dock.coat = self.coat
        dock.shell = self.shell

    def coat(self, datum, as_const=True):
        if not isinstance(datum, torch.Tensor): # numpy array
            if datum.shape==(): # scalar
                datum = torch.tensor(datum)
            else:
                datum = torch.from_numpy(datum)

        if as_const:
            datum.requires_grad = False
        else:
            datum.requires_grad = True

        if self.param['device'] == 'gpu':
            if self.param['rank']<0:
                return datum.cuda()
            else:
                return datum.to(self.device[self.param['rank']])
        elif self.param['device'] == 'cpu':
            return datum.cpu()
        else:
            raise KeyError('NEBULAE ERROR ⨷ given device should be either cpu or gpu.')

    def shell(self, datum, as_np=True):
        datum = datum.detach()
        if as_np:
            if datum.size == 1:
                datum = datum.cpu().numpy()[0]
            else:
                datum = datum.cpu().numpy()
        return datum


#!/usr/bin/env python
'''
architect
Created by Seria at 2021/5/22 11:23 AM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
from ... import dock

import numpy as np


class RNNE(dock.Craft):
    def __init__(self, in_chs, hid_chs, nlayers, voc_size, scope='RNNE'):
        super(RNNE, self).__init__(scope)
        self.emb = dock.Embed(voc_size, in_chs)
        self.rnn = dock.RNN(in_chs, hid_chs, nlayers)

    def run(self, x, h=None):
        x = self.emb(x)
        y, h = self.rnn(x, h)
        return y, h


class BiRNNE(dock.Craft):
    def __init__(self, in_chs, hid_chs, nlayers, voc_size, scope='BIRNNE'):
        super(BiRNNE, self).__init__(scope)
        self.emb = dock.Embed(voc_size, in_chs)
        self.birnn = dock.BiRNN(in_chs, hid_chs, nlayers)

    def run(self, x, h=None):
        x = self.emb(x)
        y, h = self.birnn(x, h)
        return y, h


class RNND(dock.Craft):
    def __init__(self, in_chs, hid_chs, nlayers, voc_size, scope='RNND'):
        super(RNND, self).__init__(scope)
        self.emb = dock.Embed(voc_size, in_chs)
        self.relu = dock.Relu()
        self.rnn = dock.RNN(in_chs, hid_chs, nlayers)
        self.fc = dock.Dense(hid_chs, voc_size)

    def run(self, x, h=None):
        x = self.emb(x)
        x = self.relu(x)
        y, h = self.rnn(x, h)
        y = self.fc(y)
        return y, h


# class AttnRNND(dock.Craft):
#     def __init__(self, in_chs, hid_chs, nlayers, voc_size, max_len, scope='ATTNRNND'):
#         super(AttnRNND, self).__init__(scope)
#         self.emb = dock.Embed(voc_size, in_chs)
#         self.lut = [p for p in self.emb.vars()]
#         self.relu = dock.Relu()
#         self.rnn = dock.RNN(in_chs, hid_chs, nlayers)
#         self.fc = dock.Dense(hid_chs, voc_size)
#         self.drp = dock.Dropout(0.1, dim=1)
#         self.perm = dock.Permute()
#         # self.sqsh = dock.Squash()
#         # self.expd = dock.Expand()
#
#         self.attn = dock.Dense(in_chs+hid_chs, max_len)
#         self.proj = dock.Dense(hid_chs+hid_chs, hid_chs)
#         self.dot = dock.Dot()
#         self.cat = dock.Concat()
#         self.sftm = dock.Sftm()
#
#     def run(self, x, o, h=None):
#         x = self.emb(x)
#         x = self.drp(x)
#
#         a = self.cat((x, h))
#         a = self.attn(a)
#         a = self.sftm(a)
#         _a = self.perm(a, (1, 0, 2))
#         _o = self.perm(o, (1, 0, 2))
#         a = self.dot(_a, _o, in_batch=True)
#         a = self.perm(a, (1, 0, 2))
#         h = self.cat((h, a))
#         h = self.proj(h)
#
#         x = self.relu(x)
#         y, h = self.rnn(x, h)
#         y = self.fc(y)
#         # y = self.expd(self.dot(self.sqsh(y, 0), self.perm(self.lut[0], (1,0))), 0)
#         return y, h


class AttnRNND(dock.Craft):
    def __init__(self, in_chs, hid_chs, nlayers, voc_size, max_len, scope='ATTNRNND'):
        super(AttnRNND, self).__init__(scope)
        self.nlay = nlayers
        self.hchs = hid_chs

        self.emb = dock.Embed(voc_size, in_chs)
        self.resh = dock.Reshape()
        self.perm = dock.Permute()
        self.expd = dock.Expand()
        self.sqsh = dock.Squash()
        self.relu = dock.Relu()
        self.rnn = dock.RNN(in_chs+hid_chs, hid_chs, nlayers)
        self.fc = dock.Dense(hid_chs, voc_size)
        self.do = dock.Dense(hid_chs, hid_chs)
        self.dh = dock.Dense(hid_chs, hid_chs)
        self.da = dock.Dense(hid_chs, 1)

        # self.proj = dock.Dense(in_chs+2*hid_chs, in_chs)
        self.dot = dock.Dot()
        self.cat = dock.Concat()
        self.sftm = dock.Sftm()

    def run(self, x, o, h=None):
        x = self.emb(x)
        h = self.resh(h, (self.nlay, 1, -1))

        _o = self.perm(self.do(o), (1, 0, 2)) # B x L x H*NDIR
        _h = self.perm(self.dh(h), (1, 2, 0)) # B x H*NDIR x NLAY
        # a = self.dot(_o, _h, in_batch=True) # B x L x NLAY
        a = _o + self.expd(self.sqsh(_h, -1), 1) # B x L x H*NDIR
        a = self.expd(self.sqsh(self.da(a), -1), 1) # B x 1 x L
        a = self.sftm(a)
        a = self.dot(a, _o, in_batch=True) # B x 1 x H*NDIR
        a = self.perm(a, (1, 0, 2))
        x = self.cat((x, a)) # 1 x B x 2*H*NDIR
        # x = self.proj(x)

        # x = self.relu(x)
        y, h = self.rnn(x, h)
        y = self.fc(y)
        return y, h


class LSTME(dock.Craft):
    def __init__(self, in_chs, hid_chs, nlayers, voc_size, scope='LSTME'):
        super(LSTME, self).__init__(scope)
        self.emb = dock.Embed(voc_size, in_chs)
        self.lstm = dock.LSTM(in_chs, hid_chs, nlayers)

    def run(self, x, h=None, c=None):
        x = self.emb(x)
        y, h, c = self.lstm(x, h, c)
        return y, h, c


class LSTMD(dock.Craft):
    def __init__(self, in_chs, hid_chs, nlayers, voc_size, attention=0, scope='LSTMD'):
        super(LSTMD, self).__init__(scope)
        self.emb = dock.Embed(voc_size, in_chs)
        self.relu = dock.Relu()
        self.lstm = dock.LSTM(in_chs, hid_chs, nlayers)
        self.fc = dock.Dense(hid_chs, voc_size)
        if attention>0:
            self.attn = dock.Dense(in_chs+hid_chs, attention)

    def run(self, x, h=None, c=None):
        x = self.emb(x)
        x = self.relu(x)
        y, h, c = self.lstm(x, h, c)
        y = self.fc(y)
        return y, h, c


class AttnLSTMD(dock.Craft):
    def __init__(self, in_chs, hid_chs, nlayers, voc_size, max_len, scope='ATTNLSTMD'):
        super(AttnLSTMD, self).__init__(scope)
        self.emb = dock.Embed(voc_size, in_chs)
        self.relu = dock.Relu()
        self.rnn = dock.LSTM(in_chs, hid_chs, nlayers)
        self.fc = dock.Dense(hid_chs, voc_size)
        self.perm = dock.Permute()

        self.attn = dock.Dense(in_chs+hid_chs, max_len)
        self.proj = dock.Dense(hid_chs+hid_chs, hid_chs)
        self.dot = dock.Dot()
        self.cat = dock.Concat()
        self.sftm = dock.Sftm()

    def run(self, x, o, h=None, c=None):
        x = self.emb(x)

        a = self.cat((x, c))
        a = self.attn(a)
        a = self.sftm(a)
        _a = self.perm(a, (1, 0, 2))
        _o = self.perm(o, (1, 0, 2))
        a = self.dot(_a, _o, in_batch=True)
        a = self.perm(a, (1, 0, 2))
        c = self.cat((c, a))
        c = self.proj(c)

        x = self.relu(x)
        y, h, c = self.rnn(x, h, c)
        y = self.fc(y)
        return y, h, c


#!/usr/bin/env python
'''
engine_tf
Created by Seria at 04/02/2019 4:31 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
import os
import tensorflow as tf
from tensorflow.python.distribute.values import PerReplica
from ..toolkit.utility import GPUtil

class EngineTF(object):
    '''
    Param:
    device: 'gpu' or 'cpu'
    available_gpus:
    least_mem
    '''
    def __init__(self, param):
        self.param = param
        # look for available gpu devices
        if self.param['device'].lower() == 'gpu':
            if len(self.param['avail_gpus'])==0:
                gputil = GPUtil()
                gpus = gputil.available(self.param['ngpus'], self.param['least_mem'])
                if len(gpus) == 0:
                    raise Exception('NEBULAE ERROR ⨷ no enough available gpu', gpus)
                # TODO: multi-gpu training is to be supported
            else:
                gpus = self.param['avail_gpus']
            # convert gpu list to string
            str_gpus = ','.join([str(g[0]) for g in gpus])
            # set environment variable
            os.environ['CUDA_VISIBLE_DEVICES'] = str_gpus
            phy_gpus = tf.config.experimental.list_physical_devices('GPU')
            for pg in phy_gpus:
                tf.config.experimental.set_memory_growth(pg, True)
            tf.keras.backend.set_image_data_format('channels_first')
            if self.param['rank']<=0:
                print('+' + 20 * '-' + '+')
                print('| Reside in Devices: |')
                print('+' + 70 * '-' + '+')
                for g in gpus:
                    print('| \033[1;36mGPU {:<2d}\033[0m | {:<25s} | {:>5d} MiB free out of {:<5d} MiB |'.format(
                        g[0], g[1], g[3], g[2]
                    ))
                    print('+' + 70 * '-' + '+')
        elif self.param['device'].lower() == 'cpu':
            if self.param['rank'] <= 0:
                print('+' + (23 * '-') + '+')
                print('| Reside in Device: \033[1;36mCPU\033[0m |')
                print('+' + (23 * '-') + '+')
        else:
            raise KeyError('NEBULAE ERROR ⨷ given device should be either cpu or gpu.')

        from ..astrobase import dock
        dock.coat = self.coat
        dock.shell = self.shell

    def coat(self, datum, as_const=True):
        if not (tf.is_tensor(datum) or isinstance(datum, PerReplica)): # numpy array
            datum = tf.convert_to_tensor(datum)
        if isinstance(datum, PerReplica):
            datum = tf.concat(datum.values, axis=0)

        if not as_const:
            datum = tf.Variable(datum)

        return datum

    def shell(self, datum, as_np=True):
        if isinstance(datum, PerReplica):
            datum = tf.concat(datum.values, axis=0)
        if as_np:
            datum = datum.numpy()
        return datum


#!/usr/bin/env python
'''
component
Created by Seria at 25/11/2018 2:58 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
import os
from ..toolkit.utility import cap, autoPad
from collections import namedtuple

__all__ = ('Craft', 'Pod', 'Tensor', 'coat', 'shell')


Tensor = namedtuple('tensor', ('key', 'val'))
coat = None
shell = None


class Pod():
    def __init__(self, name, comp=[], symbol=''):
        self.name = name
        if len(comp)==0: # atomic pod
            self.comp = comp
        else:
            assert len(comp)==2
            left_sym = comp[0].symbol
            right_sym = comp[1].symbol
            if left_sym == right_sym:
                if right_sym=='':
                    self.comp = comp
                else:
                    self.comp = comp[0].comp + comp[1].comp
            else:
                if left_sym=='':
                    self.comp = [comp[0]] + comp[1].comp
                elif right_sym=='':
                    self.comp = comp[0].comp + [comp[1]]
                else:
                    self.comp = comp
        self.symbol = symbol


    def __gt__(self, other):
        return Pod('', [self, other], '>')

    def __add__(self, other):
        return Pod('', [self, other], '+')

    def __sub__(self, other):
        return Pod('', [self, other], '-')

    def __mul__(self, other):
        return Pod('', [self, other], '*')

    def __matmul__(self, other):
        return Pod('', [self, other], '@')

    def __and__(self, other):
        return Pod('', [self, other], '&')

    def __or__(self, other):
        return Pod('', [self, other], '|')

    def __xor__(self, other):
        return Pod('', [self, other], '^')

    def _cap(self, scope):
        if '/' not in self.name:
            self.name = cap(self.name, scope)
        for c in self.comp:
            c._cap(scope)


core = os.environ.get('NEB_CORE', 'PYTORCH')
if core.upper() == 'TENSORFLOW':
    from .craft_tf import *
    from . import craft_tf
    __all__ += craft_tf.__all__
elif core.upper() == 'PYTORCH':
    from .craft_pt import *
    from . import craft_pt
    __all__ += craft_pt.__all__
else:
    raise ValueError('NEBULAE ERROR ⨷ %s is an unsupported core.' % core)


#!/usr/bin/env python
'''
time_machine_tf
Created by Seria at 04/02/2019 4:35 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
import tensorflow as tf
import os
from glob import glob

class TimeMachineTF(object):
    def __init__(self, param):
        '''
        Time Machine saves current states or restores saved states
        '''
        self.param = param
        self.counter = 1
        self.anchors = []

    def to(self, craft, optz=None, file='', ckpt_scope=None, frozen=False):
        assert self.param['ckpt_path'] is not None, Exception('NEBULAE ERROR ⨷ anchor location is not provided.')

        if not isinstance(craft, tf.keras.Model): # if wrapped in EMA
            craft = craft.hull

        ckpt_path = os.path.join(self.param['ckpt_path'], file)
        if os.path.isfile(ckpt_path):
            assert optz is None, Exception('NEBULAE ERROR ⨷ single checkpoint file does not contain optimizer states.')
            moment = os.path.dirname(ckpt_path)
            craft.load_weights(moment, by_name=True, skip_mismatch=not frozen)
        else:
            moment = tf.train.Checkpoint(net=craft, optimizer=optz.optz)
            ckpt_manager = tf.train.CheckpointManager(moment, ckpt_path, max_to_keep=None)
            moment.restore(ckpt_manager.latest_checkpoint)

        if self.param['rank'] <= 0:
            print('+' + ((10 + len(ckpt_path)) * '-') + '+')
            print('| Back to \033[1;34m%s\033[0m |' % ckpt_path)
            print('+' + ((10 + len(ckpt_path)) * '-') + '+')

    def drop(self, craft, optz=None, file='', save_scope=None, frozen=False):
        if self.param['rank']>0:
            return
        assert self.param['save_path'] is not None, Exception('NEBULAE ERROR ⨷ there is nowhere to drop anchor.')

        if not isinstance(craft, tf.keras.Model): # if wrapped in EMA
            craft = craft.hull

        save_path = os.path.join(self.param['save_path'], file)
        if save_path.endswith('.h5'):
            craft.save_weights(save_path)
        else:
            if optz is None:
                save_ckpt = tf.train.Checkpoint(net=craft)
            else:
                save_ckpt = tf.train.Checkpoint(net=craft, optimizer=optz.optz)
            max_anchors = self.param['max_anchors'] if self.param['max_anchors']>0 else None
            ckpt_manager = tf.train.CheckpointManager(save_ckpt, save_path, max_to_keep=max_anchors)
            ckpt_manager.save()
        print('| Anchor is dropped at \033[1;34m%s\033[0m |' % save_path)


#!/usr/bin/env python
'''
time_machine
Created by Seria at 23/12/2018 8:34 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
import os

def TimeMachine(config=None, ckpt_path=None, save_path=None, max_anchors=-1):
    rank = int(os.environ.get('RANK', -1))
    if config is None:
        param = {'ckpt_path': ckpt_path, 'save_path': save_path, 'max_anchors': max_anchors, 'rank': rank}
    else:
        config['ckpt_path'] = config.get('ckpt_path', ckpt_path)
        config['save_path'] = config.get('save_path', save_path)
        config['max_anchors'] = config.get('max_anchors', max_anchors)
        config['rank'] = config.get('rank', rank)
        param = config

    core = os.environ.get('NEB_CORE', 'PYTORCH')
    if core.upper() == 'TENSORFLOW':
        from .time_machine_tf import TimeMachineTF
        return TimeMachineTF(param)
    elif core.upper() == 'PYTORCH':
        from .time_machine_pt import TimeMachinePT
        return TimeMachinePT(param)
    else:
        raise ValueError('NEBULAE ERROR ⨷ %s is an unsupported core.' % core)


#!/usr/bin/env python
'''
utility
Created by Seria at 14/11/2018 8:33 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
import json
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy import sparse
import subprocess as subps
import h5py
import os
import io
from math import ceil
from time import sleep
from PIL import Image
import cv2
# from ..law import Constant



def autoPad(in_size: tuple, kernel: tuple, stride=1, dilation=1):
    '''
    Args
    - in_size: input size e.g. (h, w) for 2d tensor.
               fill empty tuple or None if input size can be divided by stride along all dimensions
    - kernel: kernel size
    - stride: convolution stride
    - dilation: stride in atrous convolution

    Return
    padding elements on dimensions in reverse order
    e.g. (left, right, top, bottom, front, back) for 3d tensor
    '''
    dim = len(kernel)
    if isinstance(stride, int):
        stride = dim * [stride]
    if isinstance(dilation, int):
        dilation = dim * [dilation]

    padding = []
    if in_size is None or len(in_size)==0:
        for d in range(dim - 1, -1, -1):
            margin = - stride[d] + kernel[d] + (dilation[d] - 1) * (kernel[d] - 1)
            padding.extend([margin // 2, margin - margin // 2])
    else:
        for d in range(dim-1,-1,-1):
            margin = (ceil(in_size[d] / stride[d]) - 1) * stride[d] + kernel[d] + (dilation[d] - 1) \
                                 * (kernel[d] - 1) - in_size[d]
            padding.extend([margin//2, margin-margin//2])
    return padding

def cap(key, scope):
    return scope + '/' + key

def doff(key):
    return key.split('/')[-1]

def ver2num(version, vbits=2):
    version = version.split('.')
    number = 0
    for v in version:
        v = v.split('+')
        if len(v)>1:
            v = int(v[0], 16)
        else:
            v = int(v[0])
        number = number * 10 ** vbits + v
    return number


def leftOnExceptHook(exc_type, exc_value, tb, ignored=''):
    '''
    leave particular exception messages out
    '''
    msg = ' Traceback (most recent call last):\n'
    while tb:
        filename = tb.tb_frame.f_code.co_filename
        name = tb.tb_frame.f_code.co_name
        lineno = tb.tb_lineno
        if not ignored or ignored not in filename:
            msg += '   File "%.500s", line %d, in %.500s\n' % (filename, lineno, name)
        tb = tb.tb_next

    msg += ' %s: %s\n' %(exc_type.__name__, exc_value)
    print(msg)


def hotvec2mtx(labels, nclasses, on_value=1, off_value=0):
    '''
    Args
    labels: there are three possible format as follows,
            1. list of str
            e.g. ['0,2', '1', '1,2,3']
            2. list of array
            e.g. [[0,2], [1], [1,2,3]]
            3. list of int
            e.g. [0, 2, 1]
    nclasses: number of classes
    on_value: the value on behalf of positive label
    off_value: the value on behalf of negative label

    Returns
    dense label matrix
    '''
    batch_size = labels.shape[0]
    # initialize dense labels
    dense = off_value * np.ones((batch_size * nclasses), dtype=np.float32)
    indices = []
    if isinstance(labels[0], str):
        for b in range(batch_size):
            indices += [int(s) + b * nclasses for s in labels[b].split(Constant.CHAR_SEP)]
    elif isinstance(labels[0], (list, np.ndarray)): # labels is a nested array
        for b in range(batch_size):
            indices += [l + b * nclasses for l in labels[b]]
        dense[indices] = on_value
    else:
        for b in range(batch_size):
            indices += [int(labels[b]) + b * nclasses]
        dense[indices] = on_value
    return np.reshape(dense, (batch_size, nclasses))


def mtx2hotvec(labels):
    hot = []
    row, col = labels.shape
    for y in range(row):
        vec = []
        for x in range(col):
            if labels[y, x] != 0:
                vec.append(x)
        hot.append(vec)
    return hot


def den2spa(arr):
    return sparse.coo_matrix(arr)


def spa2den(*args):
    if len(args)==1:
        return args[0].todense()
    else:
        arr = sparse.coo_matrix((args[2], (args[0], args[1])), shape=args[-1])
        return arr.todense()


def randTruncNorm(mean, std, shape, cutoff_sigma=2):
    norm = np.random.normal(0, 1, (4,)+shape).astype(np.float32)
    valid = np.logical_and(norm>-cutoff_sigma, norm<cutoff_sigma)
    indices = np.argmax(valid, 0)
    norm = np.choose(indices, norm)
    norm *= std
    norm += mean
    return norm


def available(ngpus, least_mem):
    p = subps.Popen(['nvidia-smi', 'query-gpu=name,memory.total,memory.free', '--format=csv'], stdout=subps.PIPE)
    gpu_id = 0 # next gpu we are about to check
    id_mem = [] # gpu having avialable memory greater than least requirement
    title = p.stdout.readline().decode('utf-8')
    for l in p.stdout.readlines():
        name, total, vacancy = l.decode('utf-8').split(', ')
        total = int(total.split(' ')[0])
        vacancy = int(vacancy.split(' ')[0])
        if vacancy > least_mem:
            if len(id_mem) < ngpus:
                id_mem.append((gpu_id, name, total, vacancy)) # (id, info...) of gpu
            else:
                id_mem.sort(key=lambda x: x[-1])
                if vacancy > id_mem[0][1]:
                    id_mem[0] = (gpu_id, name, total, vacancy)
        gpu_id += 1
    return id_mem

def parseConfig(config_path):
    with open(config_path, 'r') as config_file:
        return json.load(config_file)

def recordConfig(config_path, config, overwrite=True):
    while not overwrite and os.path.exists(config_path):
        config_path = config_path[:-5]+'_.json'
    with open(config_path, 'w') as config_file:
        json.dump(config, config_file, indent=4)

def _mergeFuel(src_dir, src, dst, dtype, keep_src=True):
    data = {}
    info_keys = []
    shards = len(src)
    print('+' + (23 + 2 * shards + len(dst)) * '-' + '+')
    # read
    ending_char = '\r'
    for i, f in enumerate(src):
        hdf5 = h5py.File(os.path.join(src_dir, f), 'r')
        if i == 0:
            for key in hdf5.keys():
                info_keys.append(key)
            for key in info_keys:
                if key == Constant.FRAME_KEY:
                    data[key] = 0
                else:
                    data[key] = []
        for key in info_keys:
            if key == Constant.FRAME_KEY:
                frames = hdf5[key]
                data[key] = frames if frames>data[key] else data[key]
            else:
                data[key].extend(hdf5[key][()].tolist())
        hdf5.close()
        if not keep_src:
            os.remove(os.path.join(src_dir, f))

        progress = i+1
        yellow_bar = progress * '❖-'
        space_bar = (shards - progress) * '◇-'
        if progress == shards:
            ending_char = '\n'
        print('| Merging data  \033[1;35m%s\033[0m  ⊰⟦-%s%s⟧⊱ |'
              % (dst, yellow_bar, space_bar), end=ending_char)
    # write
    hdf5 = h5py.File(os.path.join(src_dir, dst), 'w')
    for key in info_keys:
        if dtype[key].startswith('v'):  # dealing with encoded / varied data
            dt = h5py.special_dtype(vlen=dtype[key])
            hdf5.create_dataset(key, dtype=dt, data=np.array(data[key]))
        elif key == Constant.FRAME_KEY:
            hdf5[key] = data[key]
        else:
            hdf5[key] = np.array(data[key]).astype(dtype[key])
    hdf5.close()
    print('+' + (23 + 2 * shards + len(dst)) * '-' + '+')

def _fillFuel(src, key, data):
    hdf5 = h5py.File(src, 'a')
    if isinstance(data, (int, float, str)):
        data = [data]
    if isinstance(data, list):
        data = np.array(data)
    if data.dtype.kind == 'U': # convert unicode to string
        nbyte = data.dtype.descr[0][1].split('U')[-1]
        data_copy = data.copy()
        data = np.empty(data_copy.shape, dtype='|S'+nbyte)
        for idx, elm in np.ndenumerate(data_copy):
            data[idx] = elm.encode()
    if data.dtype.kind == 'S':
        sdt = h5py.special_dtype(vlen=str)
        hdf5.create_dataset(key, dtype=sdt, data=data)
    else:
        hdf5[key] = data
    hdf5.close()

def _deductFuel(src, key):
    hdf5 = h5py.File(src, 'a')
    del hdf5[key]
    hdf5.close()

def byte2arr(data_src, as_np=True):
    data_bytes = data_src.tobytes()
    if as_np:
        data_np = np.frombuffer(data_bytes, dtype='uint8')
        data_np = cv2.imdecode(data_np, -1)
        # t_ = time()
        # data_np = np.array(data_pil)  # pixels range between [0, 255]
        # print(time() - t_)
        data_np = (data_np / 255).astype(np.float32)  # pixels range between [0, 1]
        return data_np
    else:
        data_pil = Image.open(io.BytesIO(data_bytes))
        return data_pil

def curve2str(curve, divisor, span, is_global, is_elastic, x_title='x', y_title='y'):
    assert curve.ndim == 1 and span > 9 # must be a vector
    assert os.get_terminal_size().columns > span + 10 + 5 # ensure that curve is not too long to display

    line_type = {'ascent': '/', 'descent': '\\', 'vertical': '|', 'horizontal': '_'}
    if span < curve.size:
        if is_global:
            indices = [round(i * curve.size / span) for i in range(span)]
            curve = curve[indices]
            indices = [idx+1 for idx in indices]
        else:
            indices = [curve.size - i for i in range(span, 0, -1)]
            curve = curve[-span:]
    else:
        indices = [i for i in range(1, span+1)]
    y_max = curve.max()
    y_min = curve.min()
    delta = (y_max - y_min) / divisor
    if delta == 0.:
        grid = [[10 * ' ' + ' ┃ ', span * ' ', '\n'] for i in range(1, divisor + 1)]
        grid.append([10 * ' ' + ' ▲ ', span * ' ', '\n'])
    else:
        if is_elastic:
            quant = np.clip(np.floor((curve - y_min) / delta).astype(np.int8), 0, divisor-1)
            hist = np.zeros(divisor)
            qualified = np.ones(divisor, dtype=np.int8)
            increment = 1. / quant.size
            # build histogram for bins along the vertical axis
            for q in quant:
                hist[q] += increment
            ascend = np.argsort(hist)
            portion = 1. / divisor
            merged = 0
            segment = 0
            contig = 0
            sum_h = 0
            # merge rarely plotted areas
            last = -1 # where the last merged segment ends
            for k, h in enumerate(hist):
                sum_h += h
                if sum_h < portion and k < divisor-1:
                    contig += 1
                elif k == divisor-1:
                    stop = k
                    if h < portion and contig > 0:
                        stop += 1
                    if h >= portion and contig == 1:
                        contig = 0
                    qualified[k - contig: stop] *= 0
                    merged += stop - (k - contig)
                    if stop - (k - contig) > 0 and (last < 0 or last != k-contig):
                        segment += 1
                else:
                    if contig > 1:
                        stop = k
                        if h < portion:
                            stop += 1
                        qualified[k-contig : stop] *= 0
                        merged += stop - (k - contig)
                        if last < 0 or last != k - contig:
                            segment += 1
                        last = stop
                        contig = 0
                        sum_h = 0
                    elif h < portion:
                        sum_h = h
                        contig = 1
                    else:
                        contig = 0
                        sum_h = 0
            # replot curves
            delimiter = [0]
            quot = (merged - segment) % (divisor - merged)
            univ = (merged - segment) // (divisor - merged)
            qualified *= univ + 1
            if quot > 0:
                for a in ascend[-quot:]:
                    qualified[a] += 1
            m_ = -1
            _m = -1
            quant = np.zeros_like(quant)
            for k, q in enumerate(qualified):
                if q == 0:
                    m_ = k if m_<0 else m_
                    _m = k
                else:
                    if _m>=0:
                        delimiter.append((_m + 1) * delta)
                        quant[curve - y_min - delimiter[-2] > 0] = len(delimiter) - 1
                    for j in range(q):
                        delimiter.append(k * delta + (j+1) * delta/q)
                        quant[curve - y_min - delimiter[-2] > 0] = len(delimiter) - 1
                    m_ = -1
                    _m = -1
            if _m >= 0:
                assert m_ != _m
                delimiter.append((_m + 1) * delta)
                quant[curve - y_min - delimiter[-2] > 0] = len(delimiter) - 1
            delimiter = delimiter[1:]
        else:
            quant = np.round((curve - y_min) / delta).astype(np.int8)
            delimiter = [i * delta for i in range(1, divisor+1)]

        grid = [[f'{y_min + d:>10.3f} ┃ ', span * ' ', '\n'] for d in delimiter]
        grid.append([10 * ' ' + ' ▲︎ ', span * ' ', '\n'])
        # draw the curve
        for i in range(curve.size-1):
            prev = quant[i]
            curr = quant[i + 1]
            if prev > curr:
                grid[prev - 1][1] = grid[prev - 1][1][:i] + line_type['descent'] + grid[prev - 1][1][i + 1:]
                for j in range(1, prev - curr):
                    grid[prev - 1 - j][1] = grid[prev - 1 - j][1][:i] + line_type['vertical'] + grid[prev - 1 - j][1][
                                                                                                i + 1:]
            elif prev < curr:
                grid[prev][1] = grid[prev][1][:i] + line_type['ascent'] + grid[prev][1][i + 1:]
                for j in range(1, curr - prev):
                    grid[prev + j][1] = grid[prev + j][1][:i] + line_type['vertical'] + grid[prev + j][1][i + 1:]
            else:
                grid[prev][1] = grid[prev][1][:i] + line_type['horizontal'] + grid[prev][1][i + 1:]

    # initialize axis
    cstr = ''
    cstr += 10 * ' ' + ' %s\n'%y_title
    for i in range(divisor, -1, -1):
        cstr += ''.join(grid[i])
    cstr += f'{y_min:>10.3f} ┗━' + span * '━' + ' ► %s\n'%x_title
    x_domain = (10+3) * ' '
    for i in range(0, len(indices), 5):
        idx = indices[i]
        if idx<1e3:
                x_domain += f'{idx:<4d} '
        elif idx<1e4:
                x_domain += f'{idx/1e3:<3.1f}K '
        elif idx<1e6:
                x_domain += f'{round(idx/1e3):<3d}K '
        elif idx<1e7:
                x_domain += f'{idx/1e6:<3.1f}M '
        elif idx<1e9:
                x_domain += f'{round(idx/1e6):<3d}M '
    cstr += x_domain+'\n'

    return cstr



def joinImg(imgs, nrow, ncol):
    N, H, W, C = imgs.shape
    assert N == nrow*ncol, 'NEBULAE ERROR ⨷ the number of images does not match cells.'
    assert N > 1, 'NEBULAE ERROR ⨷ one image does not need to be pieced together.'
    margin_h = max(1, H//20)
    margin_w = max(1, W//20)
    canvas = np.zeros((margin_h*(nrow+1) + H*nrow, margin_w*(ncol+1) + W*ncol, C))
    for c in range(ncol):
        for r in range(nrow):
            y = margin_h*(r+1) + H*r
            x = margin_w*(c+1) + W*c
            canvas[y:y+H, x:x+W] = imgs[r*ncol + c]
    return canvas


def plotInOne(crv_names, crv_files, dst_path):
    assert len(crv_names)==len(crv_files), 'NEBULAE ERROR ⨷ the number of files does not match names.'
    assert len(crv_names) > 1, 'NEBULAE ERROR ⨷ one curve does not need to be drawn together.'
    palette = ['#F08080', '#00BFFF', '#FFFF00', '#2E8B57', '#6A5ACD', '#FFD700', '#808080']
    i = 0
    for n, f in zip(crv_names, crv_files):
        data = pd.read_csv(f, header=None).values
        plt.plot(data[:,0], data[:,1], c=palette[i % len(palette)], label=n)
        i += 1
    plt.legend()
    plt.grid(True)
    plt.savefig(dst_path)
    plt.close()



def qrGramSchmidt(A):
    """Gram-schmidt Orthogonalization"""
    Q=np.zeros_like(A)
    cnt = 0
    for a in A.T:
        u = np.copy(a)
        for i in range(0, cnt):
            u -= np.dot(np.dot(Q[:, i].T, a), Q[:, i]) # minus vector projection
        e = u / np.linalg.norm(u)  # normalization
        Q[:, cnt] = e
        cnt += 1
    R = np.dot(Q.T, A)
    return Q, R

def qrGivens(A):
    """Givens Rotation"""
    (r, c) = np.shape(A)
    Q = np.identity(r)
    R = np.copy(A)
    (rows, cols) = np.tril_indices(r, -1, c)
    for (row, col) in zip(rows, cols):
        if R[row, col] != 0:  # R[row, col]=0 -> c=1,s=0, R and Q stay unchanged
            r_ = np.hypot(R[col, col], R[row, col])  # d
            c = R[col, col]/r_
            s = -R[row, col]/r_
            G = np.identity(r)
            G[[col, row], [col, row]] = c
            G[row, col] = s
            G[col, row] = -s
            R = np.dot(G, R)  # R=G(n-1,n)*...*G(2n)*...*G(23,1n)*...*G(12)*A
            Q = np.dot(Q, G.T)  # Q=G(n-1,n).T*...*G(2n).T*...*G(23,1n).T*...*G(12).T
    return Q, R

def qrHouseholder(A):
    """Householder Reflection"""
    (r, c) = np.shape(A)
    Q = np.identity(r)
    R = np.copy(A)
    for cnt in range(r - 1):
        x = R[cnt:, cnt]
        e = np.zeros_like(x)
        e[0] = np.linalg.norm(x)
        u = x - e
        v = u / np.linalg.norm(u)
        Q_cnt = np.identity(r)
        Q_cnt[cnt:, cnt:] -= 2.0 * np.outer(v, v)
        R = np.dot(Q_cnt, R)  # R=H(n-1)*...*H(2)*H(1)*A
        Q = np.dot(Q, Q_cnt)  # Q=H(n-1)*...*H(2)*H(1)
    return Q, R



class GPUtil():
    def __init__(self):
        self.process = None # the monitor process
        self.stat = 'No statistic for now.'
        self.gpus = [] # GPU names
        with subps.Popen(['nvidia-smi', '--query-gpu=name', '--format=csv'], stdout=subps.PIPE) as p:
            _ = p.stdout.readline()
            for line in p.stdout.readlines():
                self.gpus.append(line.decode('utf-8').strip())

    def __str__(self):
        return self.stat

    def _stamp2secs(self, t):
        date, time = t.split(' ')
        h, m, s = time.split(':')
        return 3600 * int(h) + 60 * int(m) + float(s)

    def available(self, ngpus, least_mem):
        with subps.Popen(['nvidia-smi', '--query-gpu=name,memory.total,memory.free', '--format=csv'],
                        stdout=subps.PIPE) as p:
            gpu_id = 0  # next gpu we are about to check
            id_mem = []  # gpu having avialable memory greater than least requirement
            title = p.stdout.readline().decode('utf-8')
            for l in p.stdout.readlines():
                name, total, vacancy = l.decode('utf-8').strip().split(', ')
                total = int(total.split(' ')[0])
                vacancy = int(vacancy.split(' ')[0])
                if vacancy > least_mem:
                    if len(id_mem) < ngpus:
                        id_mem.append((gpu_id, name, total, vacancy))  # (id, info...) of gpu
                    else:
                        id_mem.sort(key=lambda x: x[-1])
                        if vacancy > id_mem[0][-1]:
                            id_mem[0] = (gpu_id, name, total, vacancy)
                gpu_id += 1
        return id_mem

    def monitor(self, sec=5):
        assert isinstance(sec, int), 'NEBULAE ERROR ⨷ the monitoring interval must be an integer.'
        if sec<5:
            print('NEBULAE WARNING ◘ monitor GPU too often might cause abnormal statistics.')
        self.file = open('./temp_gpu_stat.csv', 'w')
        # time limit is 12 hours
        self.process = subps.Popen(['timeout', '43200', 'nvidia-smi',
                                    '--query-gpu=timestamp,temperature.gpu,utilization.gpu,memory.used',
                                    '--format=csv', '-l', '%d'%sec],
                                   stdout=self.file, stderr=subps.PIPE)

    def status(self):
        while self.process.poll() != 0: # if monitor hasn't terminated
            subps.call(['pkill', 'nvidia-smi'])
            sleep(1)
        self.file.close()
        with open('./temp_gpu_stat.csv', 'r') as f:
            gpu_id = 0
            n = len(self.gpus)
            t_ = ''
            prev_s = [-1 for _ in range(n)]
            prev_t = [-1 for _ in range(n)]
            prev_u = [-1 for _ in range(n)]
            prev_m = [-1 for _ in range(n)]
            curr_s = [-1 for _ in range(n)]
            curr_t = [-1 for _ in range(n)]
            curr_u = [-1 for _ in range(n)]
            curr_m = [-1 for _ in range(n)]
            area_t = [0 for _ in range(n)]
            area_u = [0 for _ in range(n)]
            area_m = [0 for _ in range(n)]
            line = f.readline()
            while True:
                line = f.readline()
                if not line:
                    break
                line = line.strip().split(', ')
                if len(line)!=4 or len(line[0])!=23 or line[-1][-3:]!='MiB': # skip title line and error lines
                    print('NEBULAE WARNING ◘ GPU log file is truncated due to part of damaged numbers.')
                    break
                minute, second = line[0][-9:].split(':')
                curr_s[gpu_id] = 60 * int(minute) + float(second)
                curr_t[gpu_id] = int(line[1])
                curr_u[gpu_id] = int(line[2][:-2])
                curr_m[gpu_id] = int(line[3][:-4]) / 100 # avoid overfloating
                if t_ == '': # first line
                    t_ = line[0]
                else:
                    duration = curr_s[gpu_id] - prev_s[gpu_id]
                    if duration < 0: # cross hours
                        duration += 1800
                        duration = duration + 1800 if duration<0 else abs(duration-1800)
                    area_t[gpu_id] += duration * (curr_t[gpu_id] + prev_t[gpu_id]) / 2
                    area_u[gpu_id] += duration * (curr_u[gpu_id] + prev_u[gpu_id]) / 2
                    area_m[gpu_id] += duration * (curr_m[gpu_id] + prev_m[gpu_id]) / 2
                prev_s[gpu_id] = curr_s[gpu_id]
                prev_t[gpu_id] = curr_t[gpu_id]
                prev_u[gpu_id] = curr_u[gpu_id]
                prev_m[gpu_id] = curr_m[gpu_id]
                gpu_id = (gpu_id + 1) % n
                _t = line[0]
        t_ = self._stamp2secs(t_)
        _t = self._stamp2secs(_t)
        t = _t - t_
        t = t + 3600 * 24 if t<0 else t

        stat = '+' + 67 * '-' + '+'
        stat += '\n| GPU | ' + 10 * ' ' + 'Name' + 11 * ' ' + ' | T.Celsius | V-Util |   Memory   |\n'
        stat += '+' + 67 * '-' + '+'
        for i, g in enumerate(self.gpus):
            stat += '\n| {:^3d} | {:<25s} |  {:<4.1f}ºC   | {:4.1f}%  | {:<7.1f}MiB |'.format(
                i, g, area_t[i]/t, area_u[i]/t, area_m[i]/t*100)
        stat += '\n+' + 67 * '-' + '+'
        self.stat = stat
        os.remove('./temp_gpu_stat.csv')

if __name__=='__main__':
    print(curve2str(np.array([0.99,0.12,0.107,0.093,0.088,0.066,0.069,0.052,0.07,0.048]), 15, 70, True, True))